//  SpinningEarth-D3D.cpp
//
//       YOU DO NOT NEED TO READ THIS FILE TO UNDERSTAND
//             COMPUTER GRAPHICS IN CURVED SPACES.
//
//  CurvedSpaces-Simple includes the much simpler SceneryObjectCube
//  and SceneryObjectOctahedron, which illustrate the basic principles.

#include "SpinningEarth-D3D.hpp"
#include <assert.h>

#define RADIANS_PER_SECOND  (D3DX_PI/4)
#define EARTH_RADIUS        0.05f

//  The standard D3DXCreateSphere() is easy to use,
//  but its rectangular triangulation scheme, aligned along
//  latitude and longitude lines, is woefully inefficient,
//  especially near the poles.  Use an efficient icosahedral
//  triangulation instead, subdivided as necessary.

//  2003/12/21  Consolidate parallel arrays into a single
//  array of structures to represent the various subdivisions
//  of the triangulation.
//
//  2003/12/21  Replace longitude-latitude texture parameterization
//  with separate northern hemisphere and southern hemisphere
//  projections.  Each hemisphere projects radially onto
//  an octahedron and then orthogonally onto the xy-plane.
//  This avoids the anomalies at the north and south poles
//  as well as the problem with coordinate-wrapping.
//  Note too that the texture coordinates agree at the equator,
//  so no repetition of vertices is needed.  Conceptually
//  the sphere is being mapped onto the texture plane
//  with no rips or tears -- the only distinction between
//  the northern and southern hemispheres is which texture
//  gets applied.
//
//  2003/12/21  Replace icoshedral triangulation with octahedal one.


CSpinningEarth::CSpinningEarth(void)
{
    int i;

    itsRotationAngle = 0.0;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        itsSubdivisions[i].itsNumVertices           = 0;
        itsSubdivisions[i].itsVertices              = NULL;
        itsSubdivisions[i].itsNumFaces              = 0;
        itsSubdivisions[i].itsFaces                 = NULL;
    }

    InitOctahedron(&itsSubdivisions[0]);

    for (i = 0; i < REFINEMENT_LEVEL; i++)
        SubdivideTriangulation(&itsSubdivisions[i], &itsSubdivisions[i+1]);
}

CSpinningEarth::~CSpinningEarth(void)
{
    int i;

    for (i = 0; i <= REFINEMENT_LEVEL; i++)
    {
        SAFE_DELETE_(itsSubdivisions[i].itsVertices)
        SAFE_DELETE_(itsSubdivisions[i].itsFaces)
    }

}


void CSpinningEarth::Update(double anElapsedTime)
{
    //  To account for differing orientation conventions
    //  between Direct3D and OpenGL versions of this program,
    //  decrease itsRotationAngle instead of increasing it.

    itsRotationAngle -= RADIANS_PER_SECOND * anElapsedTime;

    if (itsRotationAngle < 0)
        itsRotationAngle += 2*D3DX_PI;
}


void CSpinningEarth::Render(
    LPDIRECT3DDEVICE7       aDevice,
    D3DXMATRIX              *aPosition,
    D3DXMATRIX              *aCameraPosition,
    LPDIRECTDRAWSURFACE7    aNorthernTexture,
    LPDIRECTDRAWSURFACE7    aSouthernTexture,
    DWORD                   aColor,
    int                     aFogRange,
    DWORD                   (*aFogFunction)(float *, float *, int, float *))
{
    D3DXMATRIX          theRotation,
                        theQuarterTurn,
                        theAttitude,
                        theWorldMatrix;
    DWORD               theFogFactor;
    float               theDistance;
    int                 theRefinementLevel,
                        i;

    //  For simplicity use a single fog factor for the whole Earth.
    theFogFactor = (*aFogFunction)( aCameraPosition->m[3],
                                    aPosition->m[3],
                                    aFogRange,
                                    &theDistance);

    try
    {
        D3DXMatrixRotationZ(&theRotation, (float) itsRotationAngle);
        D3DXMatrixRotationX(&theQuarterTurn, -D3DX_PI/2);
        D3DXMatrixMultiply(&theAttitude, &theRotation, &theQuarterTurn);
        D3DXMatrixMultiply(&theWorldMatrix, &theAttitude, aPosition);
        D3DTRY(aDevice->SetTransform(D3DTRANSFORMSTATE_WORLD, theWorldMatrix))

        D3DTRY(aDevice->SetTextureStageState(0, D3DTSS_ADDRESS, D3DTADDRESS_CLAMP));

        //  Pick an appropriate level of detail at which
        //  to render this image of the spinning Earth.
        assert(REFINEMENT_LEVEL >= 3);
        if (theDistance < 1.0)
            theRefinementLevel = 3;
        else
        if (theDistance < 1.58) //  just above pi/2
            theRefinementLevel = 2;
        else
        if (theDistance < 2.0)
            theRefinementLevel = 1;
        else
            theRefinementLevel = 0;

        for (i = 0; i < itsSubdivisions[theRefinementLevel].itsNumVertices; i++)
        {
            itsSubdivisions[theRefinementLevel].itsVertices[i].specular   = theFogFactor;
            itsSubdivisions[theRefinementLevel].itsVertices[i].color      = aColor;
        }

        D3DTRY(aDevice->SetTexture(0, aNorthernTexture))
        D3DTRY(aDevice->DrawIndexedPrimitive(
                    D3DPT_TRIANGLELIST,
                    D3DFVF_LVERTEX,
                    itsSubdivisions[theRefinementLevel].itsVertices,
                    itsSubdivisions[theRefinementLevel].itsNumVertices,
                    itsSubdivisions[theRefinementLevel].itsFaces,
                    3*itsSubdivisions[theRefinementLevel].itsNumFaces / 2,
                    0))

        D3DTRY(aDevice->SetTexture(0, aSouthernTexture))
        D3DTRY(aDevice->DrawIndexedPrimitive(
                    D3DPT_TRIANGLELIST,
                    D3DFVF_LVERTEX,
                    itsSubdivisions[theRefinementLevel].itsVertices,
                    itsSubdivisions[theRefinementLevel].itsNumVertices,
                    itsSubdivisions[theRefinementLevel].itsFaces
                      + 3*itsSubdivisions[theRefinementLevel].itsNumFaces / 2,
                    3*itsSubdivisions[theRefinementLevel].itsNumFaces / 2,
                    0))
    }
    catch (CException *anException)
    {
        MessageBox(NULL, anException->Message(), "Rendering Error in SpinningEarth", MB_OK);
        SAFE_DELETE(anException)
    }
    catch (...)
    {
        MessageBox(NULL, "Unknown exception.", "Rendering Error in SpinningEarth", MB_OK);
    }
}


//  The remaining methods implement the octahedral triangulation
//  of the sphere.  Feel free to ignore them.

void CSpinningEarth::InitOctahedron(
    Triangulation   *aTriangulation)
{
    int     i;

    static float    v[6][3] =
                    {
                        { 1.0,  0.0,  0.0}, //  equator
                        { 0.0,  1.0,  0.0},
                        {-1.0,  0.0,  0.0},
                        { 0.0, -1.0,  0.0},
                        { 0.0,  0.0,  1.0}, //  north pole
                        { 0.0,  0.0, -1.0}  //  south pole
                    };

    //  Make sure all northern faces come before all southern faces,
    //  so they may be eventually be rendered separately using
    //  different texture maps.
    static WORD     f[24] =
                    {
                        //  northern faces
                        4, 0, 1,
                        4, 1, 2,
                        4, 2, 3,
                        4, 3, 0,

                        //  southern faces
                        5, 1, 0,
                        5, 2, 1,
                        5, 3, 2,
                        5, 0, 3
                    };

    aTriangulation->itsNumVertices          = 6;
    aTriangulation->itsNumFaces             = 8;
    aTriangulation->itsVertices             = new D3DLVERTEX[6];
    aTriangulation->itsFaces                = new WORD[3*8];

    assert(aTriangulation->itsVertices != NULL);
    assert(aTriangulation->itsFaces    != NULL);

    for (i = 0; i < 6; i++)
    {
        aTriangulation->itsVertices[i].x = v[i][0] * EARTH_RADIUS;
        aTriangulation->itsVertices[i].y = v[i][1] * EARTH_RADIUS;
        aTriangulation->itsVertices[i].z = v[i][2] * EARTH_RADIUS;

//      aTriangulation->itsVertices[i].nx = v[i][0];
//      aTriangulation->itsVertices[i].ny = v[i][1];
//      aTriangulation->itsVertices[i].nz = v[i][2];

        aTriangulation->itsVertices[i].color    = 0xFFFFFFFF;
        aTriangulation->itsVertices[i].specular = 0xFFFFFFFF;

        ComputeTextureCoordinates(v[i], &aTriangulation->itsVertices[i]);
    }

    for (i = 0; i < 24; i++)
        aTriangulation->itsFaces[i] = f[i];
}


void CSpinningEarth::ComputeTextureCoordinates(
    float       v[3],
    D3DLVERTEX  *aVertex)
{
    double  x,
            y,
            z,
            theLength;

    //  Copy the vertex coordinates.
    x = v[0];
    y = v[1];
    z = v[2];

    //  Project radially onto an octahedron.
    theLength = fabs(x) + fabs(y) + fabs(z);
    if (theLength > 0)
    {
        x /= theLength;
        y /= theLength;
        z /= theLength;
    }
    else
    {
        //  Should never occur!
        aVertex->tu = aVertex->tv = 0.0;
        return;
    }

    //  Project orthogonally onto the xy-plane.
    //  (Just ignore z!)

    //  Rotate 45 degrees and rescale to [0,1].
    aVertex->tu = (float)(0.5*(x + y) + 0.5);
    aVertex->tv = (float)(0.5*(x - y) + 0.5);

    //  Account for differing orientation conventions
    //  between Direct3D and OpenGL versions of this program.
    aVertex->tu = 1.0f - aVertex->tu;
    
    //  For insurance against rounding errors...
    if (aVertex->tu < 0.0f)
        aVertex->tu = 0.0f;
    if (aVertex->tu > 1.0f)
        aVertex->tu = 1.0f;
    if (aVertex->tv < 0.0f)
        aVertex->tv = 0.0f;
    if (aVertex->tv > 1.0f)
        aVertex->tv = 1.0f;
}

void CSpinningEarth::SubdivideTriangulation(
    Triangulation *aTriangulation,
    Triangulation *aSubdivision)
{
    //  Subdivide the triangulation, replacing
    //  each old triangle with four new ones.
    //
    //    /\
    //   /__\
    //  /_\/_\
    //

    WORD        u,
                v,
                *f,
                ff[3];
    float       normal[3];
    double      theLengthSquared,
                theFactor;
    int         *theTable,
                theVertexCount,
                i,
                j,
                k;

    aSubdivision->itsNumVertices        = aTriangulation->itsNumVertices + 3*aTriangulation->itsNumFaces/2;
    aSubdivision->itsNumFaces           = 4 * aTriangulation->itsNumFaces;
    aSubdivision->itsVertices           = new D3DLVERTEX[aSubdivision->itsNumVertices];
    aSubdivision->itsFaces              = new WORD[3*aSubdivision->itsNumFaces];

    assert(aSubdivision->itsVertices != NULL);
    assert(aSubdivision->itsFaces    != NULL);

    //  Copy the vertices from the previous level.
    for (i = 0; i < aTriangulation->itsNumVertices; i++)
        aSubdivision->itsVertices[i] = aTriangulation->itsVertices[i];
    theVertexCount = aTriangulation->itsNumVertices;

    //  Create one new vertex on each edge.
    //  Make a table to index them, so the two triangles sharing
    //  an edge can share the same vertex.  (The memory for the
    //  table grows as the square of the number of old vertices.
    //  For modest triangulations this won't be a problem.
    //  For large triangulations a fancier algorithm could be used.)

    //  In spirit theTable is a 2-dimensional array.
    //  But such an array is a hassle to allocate in C/C++,
    //  so instead we just replace theTable[i][j] with
    //  theTable[i*aTriangulation->itsNumVertices + j].
    theTable = new int[aTriangulation->itsNumVertices * aTriangulation->itsNumVertices];
    assert(theTable != NULL);

    for (i = 0; i < aTriangulation->itsNumVertices; i++)
        for (j = 0; j < aTriangulation->itsNumVertices; j++)
            theTable[i*aTriangulation->itsNumVertices + j] = -1;

    for (i = 0; i < aTriangulation->itsNumFaces; i++)
        for (j = 0; j < 3; j++)
        {
            u = aTriangulation->itsFaces[3*i +   j    ];
            v = aTriangulation->itsFaces[3*i + (j+1)%3];

            if (theTable[u*aTriangulation->itsNumVertices + v] == -1)
            {
                assert(theVertexCount < aSubdivision->itsNumVertices);

                normal[0] = (aTriangulation->itsVertices[u].x + aTriangulation->itsVertices[v].x)/2;
                normal[1] = (aTriangulation->itsVertices[u].y + aTriangulation->itsVertices[v].y)/2;
                normal[2] = (aTriangulation->itsVertices[u].z + aTriangulation->itsVertices[v].z)/2;
                theLengthSquared = normal[0]*normal[0] + normal[1]*normal[1] + normal[2]*normal[2];
                if (theLengthSquared > 0)
                    theFactor = 1.0 / sqrt(theLengthSquared);
                else    //  should never occur
                    theFactor = 1.0;
                for (k = 0; k < 3; k++)
                    normal[k] = (float)(normal[k] * theFactor);
                
                aSubdivision->itsVertices[theVertexCount].x = normal[0] * EARTH_RADIUS;
                aSubdivision->itsVertices[theVertexCount].y = normal[1] * EARTH_RADIUS;
                aSubdivision->itsVertices[theVertexCount].z = normal[2] * EARTH_RADIUS;
                
                //  This program doesn't use explicit lighting,
                //  but if desired you could record the normal vector
                //  at this point.
//              aSubdivision->itsVertices[theVertexCount].nx = normal[0];
//              aSubdivision->itsVertices[theVertexCount].ny = normal[1];
//              aSubdivision->itsVertices[theVertexCount].nz = normal[2];

                aSubdivision->itsVertices[theVertexCount].color     = 0xFFFFFFFF;
                aSubdivision->itsVertices[theVertexCount].specular  = 0xFFFFFFFF;

                ComputeTextureCoordinates(normal, &aSubdivision->itsVertices[theVertexCount]);

                theTable[u*aTriangulation->itsNumVertices + v] = theVertexCount;
                theTable[v*aTriangulation->itsNumVertices + u] = theVertexCount;

                theVertexCount++;
            }
        }
    assert(theVertexCount == aSubdivision->itsNumVertices);

    //  Create the new faces.
    //  If all the old northern faces precede
    //      all the old southern faces in aTriangulation,
    //  then all the new northern faces will precede
    //      all the new southern faces in aSubdivision.
    for (i = 0; i < aTriangulation->itsNumFaces; i++)
    {
        //  The old vertices incident to this face will be f[0], f[1] and f[2].
        f = aTriangulation->itsFaces + 3*i;

        //  The new ones will be ff[0], ff[1], and ff[2].
        //  Each ff[j] lies opposite the corresponding f[j].
        for (j = 0; j < 3; j++)
            ff[j] = theTable[f[(j+1)%3]*aTriangulation->itsNumVertices + f[(j+2)%3]];

        //  Create the new faces.

        aSubdivision->itsFaces[12*i +  0] = ff[0];
        aSubdivision->itsFaces[12*i +  1] = ff[1];
        aSubdivision->itsFaces[12*i +  2] = ff[2];

        aSubdivision->itsFaces[12*i +  3] = f[0];
        aSubdivision->itsFaces[12*i +  4] = ff[2];
        aSubdivision->itsFaces[12*i +  5] = ff[1];

        aSubdivision->itsFaces[12*i +  6] = f[1];
        aSubdivision->itsFaces[12*i +  7] = ff[0];
        aSubdivision->itsFaces[12*i +  8] = ff[2];

        aSubdivision->itsFaces[12*i +  9] = f[2];
        aSubdivision->itsFaces[12*i + 10] = ff[1];
        aSubdivision->itsFaces[12*i + 11] = ff[0];
    }

    SAFE_DELETE_(theTable)
}
