//  DirichletDomain-GL.cpp
//
//       YOU DO NOT NEED TO READ THIS FILE TO UNDERSTAND
//             COMPUTER GRAPHICS IN CURVED SPACES.
//
//      It serves only to provide some interesting content
//           in the complete version of the program.
//
//
//  Constructs a Dirichlet domain for a spherical, flat, or hyperbolic
//  3-manifold given by a Matrix array.  Assumes the Matrix array
//  contains all matrices needed for the Dirichlet domain's faces.
//  For use with CurvedSpaces-Complete-GL.cpp.
//
//  This code is adapted from SphericalDD.c, which is used with
//  SphericalManifolds.c.  The algorithm in turn comes from
//  SnapPea's Dirichlet domain algorithm.  The present algorithm
//  is simpler than SnapPea's in that (1) all necessary matrices
//  are assumed to be given, and (2) roundoff errors are less of
//  a problem because even in the hyperbolic case we'll stay close
//  to the north pole and won't do matrix multiplications.
//
//  The vertex coordinates are currently found by interpolation.
//  If greater precision were ever required, they could instead
//  be found by solving the incident face equations.  In the present
//  context, the accuracy of the interpolation method is sufficient.
//
//  By Jeff Weeks, June 2001.

#include "DirichletDomain-GL.hpp"
#include "MatrixAlgebra-GL.h"
#include <math.h>
#include <assert.h>

//  We need fairly large error epsilons because each input
//  matrix is a Matrix, whose entries are floats not doubles.
#define VERTEX_EPSILON      1e-5
#define IDENTITY_EPSILON    1e-5
#define ANTIPODAL_EPSILON   1e-5

#define CUBE_SIZE           4.0

#define BEAM_WIDTH          0.02f


//  Use a winged edge data structure for the Dirichlet domain.
 
//  The values of the following two enums may be used to index
//  static arrays, so their values shouldn't be changed.
//  Because their values are 0 and 1, they may be toggled
//  with the "not" operator '!'.
typedef int WEEdgeEnd;
enum
{
    tail = 0,
    tip  = 1
};
typedef int WEEdgeSide;
enum
{
    left  = 0,
    right = 1
};


struct WEVertex
{
    //  The Direction x gives the position of the WEVertex.
    //  Its length may be arbitrary.
    Direction       x;

    //  The point p is the same as the Direction x,
    //  but is normalized to lie on the unit 3-sphere,
    //  the flat space at w == 1, or the unit hyperbolic space.
    double          p[4];

    //  Used locally in IntersectWithHalfspace().
    double          dot_product;

    //  The WEVertices are kept on a doubly-linked list.
    WEVertex        *prev,
                    *next;
};

struct WEEdge
{
    //  v[tail] and v[tip] are the vertices incident to the
    //  tail and tip, respectively, of the directed WEEdge.
    WEVertex        *v[2];

    //  e[tail][left]  is the WEEdge incident to both v[tail] and f[left].
    //  e[tail][right] is the WEEdge incident to both v[tail] and f[right].
    //  e[tip ][left]  is the WEEdge incident to both v[tip ] and f[left].
    //  e[tip ][right] is the WEEdge incident to both v[tip ] and f[right].
    WEEdge          *e[2][2];

    //  f[left] and f[right] are the faces incident to the
    //  left and right sides, respectively, of the directed WEEdge.
    WEFace          *f[2];

    //  Precompute the information needed to draw a beam in OpenGL.
    GLfloat         beam_vertices[4][3];

    //  The WEEdges are kept on a doubly-linked list.
    WEEdge          *prev,
                    *next;
};

struct WEFace
{
    //  group_element is the 4x4 matrix that takes this face's mate
    //  to this face.  In other words, this face lies in the plane which
    //  passes orthogonally through the midpoint of the segment connecting
    //  the origin to its (the origin's) image under the group_element.
    Matrix          group_element;

    //  Is this one of the temporary faces created in SetInitialCube()?
    BOOL            is_temporary;

    //  outward_normal is a Direction in E^4 orthogonal
    //  to the hyperplane spanned by this face.
    //  The notion of "orthogonality" is ALWAYS THE EUCLIDEAN NOTION,
    //  not only in the spherical case but in the flat and hyperbolic
    //  cases as well;  this is NOT the mathematically natural
    //  definition, but it keeps the code simple.
    //  The normal is "outward" in the sense that it points
    //  to the side away from the basepoint (0,0,0,1).
    //  Its length is always 1, relative to the Euclidean metric.
    //
    //  The six faces of the initial cube in SetInitialCube()
    //  get their outward_normals set manually.  For all other faces
    //  (i.e. the real faces) the outward_normal is easily computed
    //  from the group_element.
    Direction       outward_normal;
    
    //  mate is the WEFace which is identified to this face under
    //  the action of the covering transformation group.
    WEFace          *mate;

    //  some_edge is an arbitrary WEEdge incident to the given WEFace.
    WEEdge          *some_edge;
    
    //  The to_be_removed field is used locally in InstallNewFace().
    BOOL            to_be_removed;

    //  How many sides does this face have?
    int             num_sides;
    
    //  Assign a hue to each face, so that a face and its mate
    //  have the same hue.  The hue takes values in [0.0, 1.0].
    double          hue;

    //  The WEFaces are kept on a doubly-linked list.
    WEFace          *prev,
                    *next;
};

struct WEPolyhedron
{
    int             num_vertices,
                    num_edges,
                    num_faces;

    //  The following dummy nodes serve as the beginnings and ends
    //  of the doubly-linked lists of vertices, edges and faces.
    WEVertex        vertex_list_begin,
                    vertex_list_end;
    WEEdge          edge_list_begin,
                    edge_list_end;
    WEFace          face_list_begin,
                    face_list_end;
};


//  Use typedefs for memory allocation, for future flexibility.
#if 1
#define NEW_STRUCT(struct_type)     new struct_type
#define NEW_ARRAY(n, struct_type)   new struct_type[n]
#define FREE_STRUCT(s)              SAFE_DELETE(s)
#define FREE_ARRAY(s)               SAFE_DELETE_(s)
#else
#define NEW_STRUCT(struct_type)     (struct_type *) malloc((size_t) sizeof(struct_type))
#define NEW_ARRAY(n, struct_type)   (struct_type *) malloc((size_t) (n) * sizeof(struct_type))
#define FREE_STRUCT(s)              {if ((s) != NULL) {free(s); (s) = NULL;}}
#define FREE_ARRAY(s)               FREE_STRUCT(s)
#endif


//  Convenient macros for doubly-linked lists with dummy head and tail.
#define INSERT_BEFORE(new, old)     {                               \
                                        (new)->next = (old);        \
                                        (new)->prev = (old)->prev;  \
                                        (new)->prev->next = (new);  \
                                        (new)->next->prev = (new);  \
                                    }
#define INSERT_AFTER(new, old)      {                               \
                                        (new)->prev = (old);        \
                                        (new)->next = (old)->next;  \
                                        (new)->prev->next = (new);  \
                                        (new)->next->prev = (new);  \
                                    }
#define REMOVE_NODE(node)           {                                       \
                                        (node)->next->prev = (node)->prev;  \
                                        (node)->prev->next = (node)->next;  \
                                    }


CDirichletDomain::CDirichletDomain(
    int         aNumMatrices,
    Matrix      *aMatrixList,
    SpaceTypeDD aSpaceType)
{
    //  Initialize itsPolyhedron to NULL to avoid
    //  any possibility of trouble.
    itsPolyhedron = NULL;

    //  Deal with exceptional cases.
    assert(aNumMatrices > 0);
    if (aNumMatrices == 1)
    {
        MessageBox(NULL, "This space is simply connected, so the Dirichlet domain is the whole space.", "Dirichlet Domain", MB_OK);
        return;
    }
    if (aNumMatrices == 2)
    {
        MessageBox(NULL, "The Dirichlet domain for projective space is a ball with a spherical boundary.\nThe program cannot draw it because its single spherical face has no edges.", "Dirichlet Domain", MB_OK);
        return;
    }

    //  Initialize the polyhedron to be a cube with vertices
    //  at (+-CUBE_SIZE, +-CUBE_SIZE, +-CUBE_SIZE, +1)...
    MakeCube();

    //  ...then intersect it with the half spaces defined by the matrices.
    SlicePolyhedron(aNumMatrices, aMatrixList);

    //  If faces of the original cube remain, give up.
    //  This will occur for lens spaces, and may occur for other
    //  manifolds if CUBE_SIZE is too small.
    if (TemporaryFacesRemain() == TRUE)
    {
        MessageBox(NULL, "The Dirichlet domain is too big to display properly.  Either the space is infinite, or its translation distances are too long, or it is a lens space whose Dirichlet domain is a lens shaped solid with a single great circle for its edge.", "Dirichlet Domain Too Big", MB_OK);
        FreePolyhedron();
        return;
    }

    //  Assign mates by comparing group_elements.
    AssignMates();

    //  Count the vertices, edges, and faces.
    CountCells();
    
    //  Count the number of sides each face has, and check that
    //  each face and its mate have the same number of sides.
    VerifyFaces();
    
    //  Assign a consistent hue to each face and its mate.
    AssignHues();

    //  Prepare to draw the polyhedron in OpenGL.
    PrepareBeams(aSpaceType);
}

CDirichletDomain::~CDirichletDomain(void)
{
    FreePolyhedron();
}


void CDirichletDomain::FreePolyhedron(void)
{
    WEVertex    *theDeadVertex;
    WEEdge      *theDeadEdge;
    WEFace      *theDeadFace;

    if (itsPolyhedron == NULL)
        return;

    while (itsPolyhedron->vertex_list_begin.next != &itsPolyhedron->vertex_list_end)
    {
        theDeadVertex = itsPolyhedron->vertex_list_begin.next;
        REMOVE_NODE(theDeadVertex);
        FREE_STRUCT(theDeadVertex);
    }

    while (itsPolyhedron->edge_list_begin.next != &itsPolyhedron->edge_list_end)
    {
        theDeadEdge = itsPolyhedron->edge_list_begin.next;
        REMOVE_NODE(theDeadEdge);
        FREE_STRUCT(theDeadEdge);
    }

    while (itsPolyhedron->face_list_begin.next != &itsPolyhedron->face_list_end)
    {
        theDeadFace = itsPolyhedron->face_list_begin.next;
        REMOVE_NODE(theDeadFace);
        FREE_STRUCT(theDeadFace);
    }

    FREE_STRUCT(itsPolyhedron);
}


void CDirichletDomain::CenterCamera(
    Matrix  *aCameraPosition)
{
    WEFace  *theFace;

    if (itsPolyhedron == NULL)
        return;

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        //  If the camera lies beyond this face...
        if (theFace->outward_normal[0] * aCameraPosition->m[3][0]
         +  theFace->outward_normal[1] * aCameraPosition->m[3][1]
         +  theFace->outward_normal[2] * aCameraPosition->m[3][2]
         +  theFace->outward_normal[3] * aCameraPosition->m[3][3]
         > 0.0)

            //  ...apply a group element to move the camera back.
            MatrixMultiply( aCameraPosition,
                            aCameraPosition,
                            &theFace->mate->group_element);
}


void CDirichletDomain::RenderBeams(
    float   aFogFactor,
    GLuint  aTexture)
{
    WEEdge  *theEdge;
    int     i;

    static GLfloat  theBeamTextureCoordinates[4][2] = {0,0, 0,1, 4,1, 4,0};

    if (itsPolyhedron == NULL)
        return;

    glColor3f(aFogFactor, aFogFactor, aFogFactor);

    glBindTexture(GL_TEXTURE_2D, aTexture);

    glBegin(GL_QUADS);
        for (theEdge = itsPolyhedron->edge_list_begin.next;
             theEdge != &itsPolyhedron->edge_list_end;
             theEdge = theEdge->next)
        {
            for (i = 0; i < 4; i++)
            {
                glTexCoord2fv(theBeamTextureCoordinates[i]);
                glVertex3fv(theEdge->beam_vertices[i]);
            }
        }
    glEnd();
}


void CDirichletDomain::MakeCube(void)
{
    MakeEmptyPolyhedron();
    SetInitialCube();
}


void CDirichletDomain::MakeEmptyPolyhedron(void)
{
    itsPolyhedron = NEW_STRUCT(WEPolyhedron);

    itsPolyhedron->num_vertices = 0;
    itsPolyhedron->num_edges    = 0;
    itsPolyhedron->num_faces    = 0;

    itsPolyhedron->vertex_list_begin.prev   = NULL;
    itsPolyhedron->vertex_list_begin.next   = &itsPolyhedron->vertex_list_end;
    itsPolyhedron->vertex_list_end  .prev   = &itsPolyhedron->vertex_list_begin;
    itsPolyhedron->vertex_list_end  .next   = NULL;

    itsPolyhedron->edge_list_begin.prev     = NULL;
    itsPolyhedron->edge_list_begin.next     = &itsPolyhedron->edge_list_end;
    itsPolyhedron->edge_list_end  .prev     = &itsPolyhedron->edge_list_begin;
    itsPolyhedron->edge_list_end  .next     = NULL;

    itsPolyhedron->face_list_begin.prev     = NULL;
    itsPolyhedron->face_list_begin.next     = &itsPolyhedron->face_list_end;
    itsPolyhedron->face_list_end  .prev     = &itsPolyhedron->face_list_begin;
    itsPolyhedron->face_list_end  .next     = NULL;
}


void CDirichletDomain::SetInitialCube(void)
{
    //  This function assumes an empty itsPolyhedron (i.e. the
    //  WEPolyhedron data structure is allocated and the doubly linked lists
    //  are initialized, but the WEPolyhedron has no vertices, edges or faces),
    //  and sets it to be a moderately large combinatorial cube.
    //  The vertices, faces and edges of the cube are indexed as shown
    //  in the following diagram of a cut open cube.
    //
    //                        1-------5->-------3
    //                        |                 |
    //                        |                 |
    //                        2                 3
    //                        |        5        |
    //                        V                 V
    //                        |                 |
    //                        |                 |
    //      1-------2->-------5-------7->-------7-------<-3-------3
    //      |                 |                 |                 |
    //      |                 |                 |                 |
    //      ^                 ^                 ^                 ^
    //      |        2        |        1        |        3        |
    //      8                 9                11                10
    //      |                 |                 |                 |
    //      |                 |                 |                 |
    //      0-------0->-------4-------6->-------6-------<-1-------2
    //                        |                 |
    //                        |                 |
    //                        ^                 ^
    //                        |        4        |
    //                        0                 1
    //                        |                 |
    //                        |                 |
    //                        0-------4->-------2
    //                        |                 |
    //                        |                 |
    //                        8                10
    //                        |        0        |
    //                        V                 V
    //                        |                 |
    //                        |                 |
    //                        1-------5->-------3

    int         i,
                j,
                k;
    WEVertex    *theInitialVertices[8];
    WEEdge      *theInitialEdges[12];
    WEFace      *theInitialFaces[6];

    const static int evdata[12][2] =
        {
            {0, 4},
            {2, 6},
            {1, 5},
            {3, 7},
            {0, 2},
            {1, 3},
            {4, 6},
            {5, 7},
            {0, 1},
            {4, 5},
            {2, 3},
            {6, 7}
        };
    const static int eedata[12][2][2] =
        {
            {{ 8,  4}, { 9,  6}},
            {{ 4, 10}, { 6, 11}},
            {{ 5,  8}, { 7,  9}},
            {{10,  5}, {11,  7}},
            {{ 0,  8}, { 1, 10}},
            {{ 8,  2}, {10,  3}},
            {{ 9,  0}, {11,  1}},
            {{ 2,  9}, { 3, 11}},
            {{ 4,  0}, { 5,  2}},
            {{ 0,  6}, { 2,  7}},
            {{ 1,  4}, { 3,  5}},
            {{ 6,  1}, { 7,  3}}
        };
    const static int efdata[12][2] =
        {
            {2, 4},
            {4, 3},
            {5, 2},
            {3, 5},
            {4, 0},
            {0, 5},
            {1, 4},
            {5, 1},
            {0, 2},
            {2, 1},
            {3, 0},
            {1, 3}
        };
    const static int fdata[6] = {4, 6, 0, 1, 0, 2};
    const static Direction normaldata[6] =
        {
            {-1.0,  0.0,  0.0, -4.0},
            {+1.0,  0.0,  0.0, -4.0},
            { 0.0, -1.0,  0.0, -4.0},
            { 0.0, +1.0,  0.0, -4.0},
            { 0.0,  0.0, -1.0, -4.0},
            { 0.0,  0.0, +1.0, -4.0}
        };


    //  If we were keeping track of the number of vertices, edges and faces,
    //  we'd want to set the counts here.  But we're not, so we won't.
    //
    //  itsPolyhedron->num_vertices = 8;
    //  itsPolyhedron->num_edges    = 12;
    //  itsPolyhedron->num_faces    = 6;


    for (i = 0; i < 8; i++)
    {
        theInitialVertices[i] = NEW_STRUCT(WEVertex);
        INSERT_BEFORE(theInitialVertices[i], &itsPolyhedron->vertex_list_end);
    }

    for (i = 0; i < 12; i++)
    {
        theInitialEdges[i] = NEW_STRUCT(WEEdge);
        INSERT_BEFORE(theInitialEdges[i], &itsPolyhedron->edge_list_end);
    }

    for (i = 0; i < 6; i++)
    {
        theInitialFaces[i] = NEW_STRUCT(WEFace);
        INSERT_BEFORE(theInitialFaces[i], &itsPolyhedron->face_list_end);
    }

    for (i = 0; i < 8; i++)
    {
        theInitialVertices[i]->x[0] = (i & 4) ? CUBE_SIZE : -CUBE_SIZE;
        theInitialVertices[i]->x[1] = (i & 2) ? CUBE_SIZE : -CUBE_SIZE;
        theInitialVertices[i]->x[2] = (i & 1) ? CUBE_SIZE : -CUBE_SIZE;
        theInitialVertices[i]->x[3] = 1.0;
    }

    for (i = 0; i < 12; i++)
    {
        for (j = 0; j < 2; j++)          // j = tail, tip
            theInitialEdges[i]->v[j] = theInitialVertices[evdata[i][j]];

        for (j = 0; j < 2; j++)          // j = tail, tip
            for (k = 0; k < 2; k++)      // k = left, right
                theInitialEdges[i]->e[j][k] = theInitialEdges[eedata[i][j][k]];

        for (j = 0; j < 2; j++)          // j = left, right
            theInitialEdges[i]->f[j] = theInitialFaces[efdata[i][j]];
    }

    for (i = 0; i < 6; i++)
    {
        for (j = 0; j < 4; j++)
            for (k = 0; k < 4; k++)
                theInitialFaces[i]->group_element.m[j][k] = 0.0;

        for (j = 0; j < 4; j++)
            theInitialFaces[i]->outward_normal[j] = normaldata[i][j];

        theInitialFaces[i]->is_temporary    = TRUE;
        theInitialFaces[i]->mate            = NULL;
        theInitialFaces[i]->some_edge       = theInitialEdges[fdata[i]];
        theInitialFaces[i]->num_sides       = 0;
        theInitialFaces[i]->hue             = 0.0;
    }
}


void CDirichletDomain::SlicePolyhedron(
    int     aNumMatrices,
    Matrix  *aMatrixList)
{
    //  Intersect itsPolyhedron with the halfspace defined
    //  by each D3DXMATRIX on aMatrixList.

    int     i;
    double  theWorstError;

    theWorstError = 0.0;

    for (i = 0; i < aNumMatrices; i++)
        IntersectWithHalfspace(&aMatrixList[i], &theWorstError);

    //  The scale at which vertices are getting merged should be
    //  distinctly smaller than the VERTEX_EPSILON at which merging
    //  is possible.  Alas the hyperbolic examples routinely fail
    //  this test.

//  if (theWorstError * 100 > VERTEX_EPSILON)
//      MessageBox(NULL, "Numerical accuracy is poor.", "Poor Precision", MB_OK);
}


void CDirichletDomain::IntersectWithHalfspace(
    Matrix  *aMatrix,
    double  *aWorstError)
{
    //  Intersect itsPolyhedron with the halfspace defined by aMatrix.

    Direction   theNormal,
                theMidpoint;
    double      theHorizontalComponent,
                theLength;
    BOOL        theCutIsNontrivial;
    WEVertex    *theVertex;
    WEEdge      *theEdge;
    int         i;
    double      t;
    Direction   theCutPoint;
    WEFace      *theFace;

    //  Ignore the identity matrix.
    if (MatrixIsIdentity(aMatrix) == TRUE)
        return;

    //  Ignore the antipodal matrix.
    if (aMatrix->m[3][3] < -1.0 + ANTIPODAL_EPSILON)
        return;

    //  The desired half space is the perpendicular bisector
    //  of the segment running from (0,0,0,1) to its image
    //  under the action of aMatrix.  In all three curvatures
    //  (spherical, flat, and hyperbolic) the average of the
    //  two endpoints gives the correct Direction of the midpoint
    //  (its length will be too long in hyperbolic space
    //  and too short in spherical space, but that's OK).
    //  In all three spaces we represent the hyperplane
    //  by it's Euclidean normal vector.  This is mathematically
    //  unnatural, but saves us the trouble later on of worrying
    //  about what space we're in.

    theMidpoint[0] = (aMatrix->m[3][0] + 0.0)/2;
    theMidpoint[1] = (aMatrix->m[3][1] + 0.0)/2;
    theMidpoint[2] = (aMatrix->m[3][2] + 0.0)/2;
    theMidpoint[3] = (aMatrix->m[3][3] + 1.0)/2;

    theHorizontalComponent = sqrt(theMidpoint[0]*theMidpoint[0]
                                + theMidpoint[1]*theMidpoint[1]
                                + theMidpoint[2]*theMidpoint[2]);

    assert(theHorizontalComponent > 0.0);

    theNormal[0] = theMidpoint[3] * (theMidpoint[0] / theHorizontalComponent);
    theNormal[1] = theMidpoint[3] * (theMidpoint[1] / theHorizontalComponent);
    theNormal[2] = theMidpoint[3] * (theMidpoint[2] / theHorizontalComponent);
    theNormal[3] = -theHorizontalComponent;

    //  Normalize theNormal to unit length to facilitate
    //  uniform error estimates below.
    theLength = sqrt( theNormal[0]*theNormal[0]
                    + theNormal[1]*theNormal[1]
                    + theNormal[2]*theNormal[2]
                    + theNormal[3]*theNormal[3]);
    assert(theLength > 0.0);
    theNormal[0] /= theLength;
    theNormal[1] /= theLength;
    theNormal[2] /= theLength;
    theNormal[3] /= theLength;

    //  Compute the dot product of theNormal with each vertex.
    //  The result will be negative, zero, or positive
    //  according to whether the vertex lies within the halfspace,
    //  on the boundary, or outside the halfspace.
    //  Because we're using the Euclidean normal, we use the
    //  plain old Euclidean dot product, no matter what the SpaceType.
    //
    //  If the dot_product differs from zero only by roundoff error,
    //  set it to be exactly zero.

    theCutIsNontrivial = FALSE;

    for (theVertex = itsPolyhedron->vertex_list_begin.next;
         theVertex != &itsPolyhedron->vertex_list_end;
         theVertex = theVertex->next)
    {
        theVertex->dot_product = theVertex->x[0] * theNormal[0]
                               + theVertex->x[1] * theNormal[1]
                               + theVertex->x[2] * theNormal[2]
                               + theVertex->x[3] * theNormal[3];

        if (fabs(theVertex->dot_product) < VERTEX_EPSILON)
        {
            if (*aWorstError < fabs(theVertex->dot_product))
                *aWorstError = fabs(theVertex->dot_product);
            theVertex->dot_product = 0.0;
        }

        if (theVertex->dot_product > 0.0)
            theCutIsNontrivial = TRUE;
    }

    //  If the slicing plane fails to intersect the polyhedron,
    //  there is no cutting to be done.
    if (theCutIsNontrivial == FALSE)
        return;

    //  Whenever the slicing plane crosses an edge,
    //  introduce a new vertex at the cut point.
    for (theEdge = itsPolyhedron->edge_list_begin.next;
         theEdge != &itsPolyhedron->edge_list_end;
         theEdge = theEdge->next)
    {
        if
        (
            (   theEdge->v[tail]->dot_product > 0.0
             && theEdge->v[tip ]->dot_product < 0.0)
         ||
            (   theEdge->v[tail]->dot_product < 0.0
             && theEdge->v[tip ]->dot_product > 0.0)
        )
        {
            //  A minute's thought shows that a simple linear
            //  interpolation will do the trick.

            t = (            0.0              - theEdge->v[tail]->dot_product) /
                (theEdge->v[tip]->dot_product - theEdge->v[tail]->dot_product);

            for (i = 0; i < 4; i++)
                theCutPoint[i] = (1.0 - t) * theEdge->v[tail]->x[i]
                               +      t    * theEdge->v[tip ]->x[i];

            SplitEdge(theEdge, theCutPoint);
        }
    }

    //  Whenever the slicing plane crosses a face,
    //  introduce a new edge along the cut line.

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        CutFaceIfNecessary(theFace);

    //  If we're worried about roundoff error creating
    //  a cut locus that's something other than a single circle,
    //  here's the place to verify that everything's OK.
    //  The code from
    //  CheckTopologyOfCut() in the file MatricesToDirichletDomain.c
    //  in the Cosmic Topology project or
    //  check_topology_of_cut() in the file Dirichlet_construction.c
    //  in SnapPea could be adapted to this purpose.
    //
    //  CODE NOT INSTALLED (It isn't necessary in this program.)

    //  Delete the vertices, edges and faces that are getting
    //  removed, and install the new face.
    InstallNewFace(aMatrix, theNormal);
}


void CDirichletDomain::SplitEdge(
    WEEdge      *anOldEdge,
    Direction   aCutPoint)
{
    //  Introduce a new vertex in anOldEdge at aCutPoint.
    //
    //              before                  after
    //
    //              \   /                   \   /
    //               \ /                     \ /
    //                |                       |
    //                |                 old   |
    //                |                 edge  ^
    //                |                       |
    //          old   |                       |  new
    //          edge  ^                       o vertex
    //                |                       |
    //                |                 new   |
    //                |                 edge  ^
    //                |                       |
    //                |                       |
    //               / \                     / \ 
    //              /   \                   /   \ 

    WEEdge      *theNewEdge,
                *theLeftNeighbor,
                *theRightNeighbor;
    WEVertex    *theNewVertex;
    int         i;

    //  Allocate space for the theNewEdge and theNewVertex.

    theNewEdge      = NEW_STRUCT(WEEdge);
    theNewVertex    = NEW_STRUCT(WEVertex);

    //  Set the fields of the theNewEdge.

    theNewEdge->v[tail] = anOldEdge->v[tail];
    theNewEdge->v[tip]  = theNewVertex;

    theNewEdge->e[tail][left]  = anOldEdge->e[tail][left];
    theNewEdge->e[tail][right] = anOldEdge->e[tail][right];
    theNewEdge->e[tip ][left]  = anOldEdge;
    theNewEdge->e[tip ][right] = anOldEdge;

    theNewEdge->f[left]  = anOldEdge->f[left];
    theNewEdge->f[right] = anOldEdge->f[right];

    //  Adjust the fields of the anOldEdge.

    anOldEdge->v[tail] = theNewVertex;

    anOldEdge->e[tail][left]  = theNewEdge;
    anOldEdge->e[tail][right] = theNewEdge;

    //  Adjust the fields of the other two edges that "see" theNewEdge.

    theLeftNeighbor = theNewEdge->e[tail][left];
    if (theLeftNeighbor->v[tip] == theNewEdge->v[tail])
        theLeftNeighbor->e[tip][left] = theNewEdge;
    else
    if (theLeftNeighbor->v[tail] == theNewEdge->v[tail])
        theLeftNeighbor->e[tail][right] = theNewEdge;
    else
        assert(FALSE);

    theRightNeighbor = theNewEdge->e[tail][right];
    if (theRightNeighbor->v[tip] == theNewEdge->v[tail])
        theRightNeighbor->e[tip][right] = theNewEdge;
    else
    if (theRightNeighbor->v[tail] == theNewEdge->v[tail])
        theRightNeighbor->e[tail][left] = theNewEdge;
    else
        assert(FALSE);

    //  Set the fields for the new vertex.

    for (i = 0; i < 4; i++)
        theNewVertex->x[i] = aCutPoint[i];

    theNewVertex->dot_product = 0.0;

    //  Insert theNewEdge and theNewVertex into their respective lists.
    //  Insert theNewEdge into the list just before anOldEdge,
    //  so the loop in CutEdges() doesn't see it.

    INSERT_BEFORE(theNewEdge, anOldEdge);
    INSERT_BEFORE(theNewVertex, anOldEdge->v[tip]);
}


void CDirichletDomain::CutFaceIfNecessary(
    WEFace  *aFace)
{
    //  SplitEdge() has already introduced new vertices
    //  where the slicing plane cut through old edges.
    //  aFace should have either 0, 1 or 2 vertices with dot_product 0.
    //  If it has 0 or 1, do nothing.
    //  If it has 2
    //      if the vertices are adjacent, do nothing.
    //      if the vertices are not adjacent,
    //          split the face with a new edge connecting the two vertices.
    //  If it has 3 or more,
    //      report an error (due to roundoff error).

    int     theCount;
    WEEdge  *theEdge,
            *theEdgeBefore[2],
            *theEdgeAfter[2],
            *theNewEdge;
    WEFace  *theNewFace;
    int     i;

    //  To simplify the subsequent code, reorient the WEEdges
    //  so all are directed counterclockwise around the face.
    AllEdgesCounterclockwise(aFace);

    //  Count the number of vertices with dot_product 0.
    //  Report an error if more than two are found.
    theCount = 0;
    theEdge = aFace->some_edge;
    do
    {
        if (theEdge->v[tip]->dot_product == 0.0)
        {
            //  If we've already found two new vertices incident
            //  to aFace, and this is the third one, then something
            //  has gone terribly wrong.  Most likely this will
            //  never happen, but if, say, roundoff error has caused
            //  the slicing plane to intersect aFace in some impossible
            //  way, we want to flag the error and abort.
            assert(theCount < 2);

            //  Note which edges are adjacent to the new vertex.
            theEdgeBefore[theCount] = theEdge;
            theEdgeAfter [theCount] = theEdge->e[tip][left];
            theCount++;
        }
        theEdge = theEdge->e[tip][left];
    } while (theEdge != aFace->some_edge);

    //  If there are fewer than two vertices with dot_product 0,
    //  then there is no need to cut aFace.
    if (theCount < 2)
        return;

    //  If the two vertices are adjacent,
    //  then there is no need to cut aFace.
    if (theEdgeBefore[0] == theEdgeAfter[1]
     || theEdgeBefore[1] == theEdgeAfter[0])
        return;

    //  Cut the aFace in two by introducing a new edge connecting
    //  the two new vertices.
    //
    //  Before:
    //                /               \ 
    //       theEdgeBefore[1]          \ 
    //              /                   \_
    //            |/_                   |\  theEdgeAfter[0]
    //            /                       \ 
    //           /                         \ 
    //          /           aFace           \ 
    //          \                           /
    //           \                         /
    //            \                      _/
    //            _\|                    /| theEdgeBefore[0]
    //              \                   /
    //       theEdgeAfter[1]           /
    //                \               /
    //
    //
    //  After:
    //                /               \ 
    //       theEdgeBefore[1]          \ 
    //              /                   \_
    //            |/_                   |\  theEdgeAfter[0]
    //            /      theNewFace       \ 
    //           /                         \ 
    //          /____________/______________\ 
    //          \            \ theNewEdge   /
    //           \                         /
    //            \        aFace         _/
    //            _\|                    /| theEdgeBefore[0]
    //              \                   /
    //       theEdgeAfter[1]           /
    //                \               /

    //  Allocate the theNewEdge and theNewFace.

    theNewEdge = NEW_STRUCT(WEEdge);
    theNewFace = NEW_STRUCT(WEFace);

    //  Install the theNewEdge and theNewFace.

    theNewEdge->v[tail] = theEdgeBefore[0]->v[tip];
    theNewEdge->v[tip]  = theEdgeBefore[1]->v[tip];

    theNewEdge->e[tail][left]   = theEdgeBefore[0];
    theNewEdge->e[tail][right]  = theEdgeAfter[0];
    theNewEdge->e[tip] [left]   = theEdgeAfter[1];
    theNewEdge->e[tip] [right]  = theEdgeBefore[1];

    theEdgeBefore[0]->e[tip] [left] = theNewEdge;
    theEdgeAfter [0]->e[tail][left] = theNewEdge;
    theEdgeBefore[1]->e[tip] [left] = theNewEdge;
    theEdgeAfter [1]->e[tail][left] = theNewEdge;

    for (theEdge = theEdgeAfter[0]; theEdge != theNewEdge; theEdge = theEdge->e[tip][left])
        theEdge->f[left] = theNewFace;

    theNewEdge->f[left]     = aFace;
    theNewEdge->f[right]    = theNewFace;

    theNewFace->some_edge = theNewEdge;
    aFace     ->some_edge = theNewEdge;

    theNewFace->group_element   = aFace->group_element;
    theNewFace->is_temporary    = aFace->is_temporary;
    for (i = 0; i < 4; i++)
        theNewFace->outward_normal[i] = aFace->outward_normal[i];
    theNewFace->mate            = NULL;
    theNewFace->num_sides       = 0;
    theNewFace->hue             = 0.0;

    //  We can safely insert theNewFace before aFace,
    //  because it doesn't need to get checked.
    //  theNewEdge may go anywhere.

    INSERT_BEFORE(theNewFace, aFace);
    INSERT_BEFORE(theNewEdge, theEdgeBefore[0]);
}


void CDirichletDomain::AllEdgesCounterclockwise(
    WEFace  *aFace)
{
    WEEdge  *theEdge;

    theEdge = aFace->some_edge;
    do
    {
        if (theEdge->f[left] != aFace)
            RedirectEdge(theEdge);
        theEdge = theEdge->e[tip][left];
    } while (theEdge != aFace->some_edge);
}


void CDirichletDomain::RedirectEdge(
    WEEdge  *anEdge)
{
    WEVertex    *theTempVertex;
    WEEdge      *theTempEdge;
    WEFace      *theTempFace;

    //  Swap the tip and tail vertices.
    theTempVertex   = anEdge->v[tail];
    anEdge->v[tail] = anEdge->v[tip];
    anEdge->v[tip]  = theTempVertex;

    //  Swap the tail-left and tip-right edges.
    theTempEdge             = anEdge->e[tail][left];
    anEdge->e[tail][left]   = anEdge->e[tip][right];
    anEdge->e[tip][right]   = theTempEdge;

    //  Swap the tail-right and tip-left edges.
    theTempEdge             = anEdge->e[tail][right];
    anEdge->e[tail][right]  = anEdge->e[tip][left];
    anEdge->e[tip][left]    = theTempEdge;

    //  Swap the left and right faces.
    theTempFace     = anEdge->f[left];
    anEdge->f[left] = anEdge->f[right];
    anEdge->f[right]= theTempFace;
}


void CDirichletDomain::InstallNewFace(
    Matrix      *aMatrix,
    Direction   aNormal)
{
    //  The overall plan is to remove those faces, edges and vertices
    //  (in that order) that have been cut off by the slicing plane,
    //  and install the theNewFace.

    WEFace          *theFace,
                    *theNewFace;
    WEEdge          *theEdge;
    WEVertex        *theVertex;
    WEFace          *theDeadFace;
    WEEdge          *theDeadEdge;
    WEVertex        *theDeadVertex;
    WEVertex        *theEndVertex;
    WEEdge          *theLeftEdge,
                    *theRightEdge;
    int             i;

    //  Create the new face.

    theNewFace = NEW_STRUCT(WEFace);
    theNewFace->group_element   = *aMatrix;
    theNewFace->is_temporary    = FALSE;
    for (i = 0; i < 4; i++)
        theNewFace->outward_normal[i] = aNormal[i];
    theNewFace->mate            = NULL;
    theNewFace->num_sides       = 0;
    theNewFace->hue             = 0.0;

    //  We'll want to remove precisely those faces that are incident to
    //  at least one positive vertex.  To find them, we'll scan the list
    //  of edges, and whenever an edge is incident to at least one positive
    //  vertex, we'll set the neighboring faces' to_be_removed fields to TRUE.
    //  (This is easier than traversing the perimeter of each face explicitly,
    //  worrying about the directions of the edges.)

    //  Initialize all faces' to_be_removed fields to FALSE.

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        theFace->to_be_removed = FALSE;

    //  Scan the edge list to locate edges -- and hence faces -- which
    //  are incident to at least one positive vertex.

    for (theEdge = itsPolyhedron->edge_list_begin.next;
         theEdge != &itsPolyhedron->edge_list_end;
         theEdge = theEdge->next)

        if (theEdge->v[tail]->dot_product > 0.0
         || theEdge->v[tip] ->dot_product > 0.0)
        {
            theEdge->f[left] ->to_be_removed = TRUE;
            theEdge->f[right]->to_be_removed = TRUE;
        }

    //  Set the theEdge->f[] fields for edges which will be incident to the
    //  theNewFace.  Also make sure the theNewFace's some_edge field sees
    //  an appropriate edge.  (Note that it's easier to keep setting
    //  theNewFace->some_edge over and over than to check whether it's
    //  already been set.)

    for (theEdge = itsPolyhedron->edge_list_begin.next;
         theEdge != &itsPolyhedron->edge_list_end;
         theEdge = theEdge->next)

        if (theEdge->v[tail]->dot_product == 0.0
         && theEdge->v[tip] ->dot_product == 0.0)

            for (i = 0; i < 2; i++)      // i = left, right

                if (theEdge->f[i]->to_be_removed == TRUE)
                {
                    theEdge->f[i] = theNewFace;
                    theNewFace->some_edge = theEdge;
                }

    //  Delete the faces that are being removed.

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        if (theFace->to_be_removed == TRUE)
        {
            //  Remove theFace from the doubly-linked list.  First set
            //  theFace to theFace->prev so the loop will continue correctly.
            theDeadFace = theFace;
            theFace     = theFace->prev;
            REMOVE_NODE(theDeadFace);
            FREE_STRUCT(theDeadFace);
        }

    //  Delete the edges that are being removed.

    for (theEdge = itsPolyhedron->edge_list_begin.next;
         theEdge != &itsPolyhedron->edge_list_end;
         theEdge = theEdge->next)

        if (theEdge->v[tail]->dot_product > 0.0
         || theEdge->v[tip] ->dot_product > 0.0)
        {
            //  If theEdge is incident to a zero vertex,
            //  fix up the neighbors' edge->e[][] fields.

            //  Is there a zero vertex at the tip or tail?
            if (theEdge->v[tail]->dot_product == 0.0
             || theEdge->v[tip] ->dot_product == 0.0)
            {
                //  Who are the neighbors?
                if (theEdge->v[tip]->dot_product == 0.0)
                {
                    theEndVertex = theEdge->v[tip];
                    theLeftEdge  = theEdge->e[tip][left];
                    theRightEdge = theEdge->e[tip][right];
                }
                else
                {
                    theEndVertex = theEdge->v[tail];
                    theLeftEdge  = theEdge->e[tail][right];
                    theRightEdge = theEdge->e[tail][left];
                }

                //  Let theLeftEdge see theRightEdge.
                if (theLeftEdge->v[tip] == theEndVertex)
                    theLeftEdge->e[tip][right] = theRightEdge;
                else
                {
                    assert(theLeftEdge->v[tail] == theEndVertex);
                    theLeftEdge->e[tail][left] = theRightEdge;
                }

                //  Let theRightEdge see theLeftEdge.
                if (theRightEdge->v[tip] == theEndVertex)
                    theRightEdge->e[tip][left] = theLeftEdge;
                else
                {
                    assert(theRightEdge->v[tail] == theEndVertex);
                    theRightEdge->e[tail][right] = theLeftEdge;
                }
            }

            //  Remove theEdge from the doubly-linked list.  First set
            //  theEdge to theEdge->prev so the loop will continue correctly.
            theDeadEdge = theEdge;
            theEdge     = theEdge->prev;
            REMOVE_NODE(theDeadEdge);
            FREE_STRUCT(theDeadEdge);
        }

    //  Delete the vertices that are being removed.

    for (theVertex = itsPolyhedron->vertex_list_begin.next;
         theVertex != &itsPolyhedron->vertex_list_end;
         theVertex = theVertex->next)

        if (theVertex->dot_product > 0.0)
        {
            //  Remove theVertex from the doubly-linked list.  First set
            //  theVertex to theVertex->prev so the loop will continue correctly.
            theDeadVertex   = theVertex;
            theVertex       = theVertex->prev;
            REMOVE_NODE(theDeadVertex);
            FREE_STRUCT(theDeadVertex);
        }

    //  Install the theNewFace in the doubly-linked list.

    INSERT_BEFORE(theNewFace, &itsPolyhedron->face_list_end);
}


BOOL CDirichletDomain::TemporaryFacesRemain(void)
{
    //  No faces of the original temporary cube should remain.
    //  This test will surely fail for lens spaces.
    //  If it fails for other spaces, you may need to recompile
    //  with a larger CUBE_SIZE.

    WEFace  *theFace;

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        if (theFace->is_temporary == TRUE)

            return TRUE;

    return FALSE;
}


void CDirichletDomain::AssignMates(void)
{
    WEFace  *theFace,
            *theMate;
    BOOL    theMateWasFound;
    Matrix  theProduct;

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)
    {
        //  This condition should have already been checked.
        assert(theFace->is_temporary == FALSE);

        //  Find the unique mate for theFace.

        theMateWasFound = FALSE;

        for (theMate = itsPolyhedron->face_list_begin.next;
             theMate != &itsPolyhedron->face_list_end;
             theMate = theMate->next)
        {
            MatrixMultiply( &theProduct,
                            &theFace->group_element,
                            &theMate->group_element);
            if (MatrixIsApproximatelyIdentity(&theProduct, IDENTITY_EPSILON))
            {
                assert(theMateWasFound == FALSE);
                theFace->mate   = theMate;
                theMateWasFound = TRUE;
            }
        }

        assert(theMateWasFound == TRUE);
    }
}


void CDirichletDomain::CountCells(void)
{
    WEVertex    *aVertex;
    WEEdge      *anEdge;
    WEFace      *aFace;

    //  The counts were intialized in MakeEmptyPolyhedron(),
    //  but we'll reinitialize them here just for good form.
    itsPolyhedron->num_vertices = 0;
    itsPolyhedron->num_edges    = 0;
    itsPolyhedron->num_faces    = 0;

    //  Count the vertices.
    for (aVertex = itsPolyhedron->vertex_list_begin.next;
         aVertex != &itsPolyhedron->vertex_list_end;
         aVertex = aVertex->next)
        itsPolyhedron->num_vertices++;

    //  Count the edges.
    for (anEdge = itsPolyhedron->edge_list_begin.next;
         anEdge != &itsPolyhedron->edge_list_end;
         anEdge = anEdge->next)
        itsPolyhedron->num_edges++;

    //  Count the faces.
    for (aFace = itsPolyhedron->face_list_begin.next;
         aFace != &itsPolyhedron->face_list_end;
         aFace = aFace->next)
        itsPolyhedron->num_faces++;

    //  Check the Euler number.
    assert(   itsPolyhedron->num_vertices
            - itsPolyhedron->num_edges
            + itsPolyhedron->num_faces == 2);
}


void CDirichletDomain::VerifyFaces(void)
{
    WEFace  *theFace;
    WEEdge  *theEdge;

    //  Initialize each face->num_sides to 0.

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        theFace->num_sides = 0;

    //  Add the contribution of each edge to the adjacent face->num_sides.

    for (theEdge = itsPolyhedron->edge_list_begin.next;
         theEdge != &itsPolyhedron->edge_list_end;
         theEdge = theEdge->next)
    {
        theEdge->f[left ]->num_sides++;
        theEdge->f[right]->num_sides++;
    }

    //  Check that each face and its mate have the same num_sides.

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        assert(theFace->num_sides == theFace->mate->num_sides);
}


void CDirichletDomain::AssignHues(void)
{
    //  If we eventually find ourselves working with polyhedra
    //  with large numbers of faces, we can sort the faces by size
    //  and use SnapPea's scheme for assigning to the largest faces
    //  hues with good contrast.  For now let's just assign arbitrary hues.

    WEFace  *theFace;
    int     theCount;

    //  Set each hue field to -1.0 to indicate that it has
    //  not yet been assigned.

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        theFace->hue = -1.0;
    
    //  Go through and assign integer hues.

    theCount = 0;

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        if (theFace->hue == -1.0)
        {
            theFace->hue = (double) theCount;
            theFace->mate->hue = theFace->hue;
            theCount++;
        }
    
    //  Now convert each integer hue to a real-valued hue,
    //  with values spread evenly over the range [0.0, 1.0).

    for (theFace = itsPolyhedron->face_list_begin.next;
         theFace != &itsPolyhedron->face_list_end;
         theFace = theFace->next)

        theFace->hue /= (double) theCount;
}


void CDirichletDomain::PrepareBeams(
    SpaceTypeDD aSpaceType)
{
    WEVertex    *theVertex;
    WEEdge      *theEdge;
    int         i;

    //  First normalize all the vertices to lie in their respective
    //  spaces, i.e. on the unit sphere, the flat space at height
    //  w == 1, or the unit hyperbolic space, as appropriate.

    for (theVertex = itsPolyhedron->vertex_list_begin.next;
         theVertex != &itsPolyhedron->vertex_list_end;
         theVertex = theVertex->next)
    {
        for (i = 0; i < 4; i++)
            theVertex->p[i] = theVertex->x[i];

        Normalize(theVertex->p, aSpaceType);
    }

    //  Locate the beam_vertices along the Dirichlet domain's edges.

    for (theEdge = itsPolyhedron->edge_list_begin.next;
         theEdge != &itsPolyhedron->edge_list_end;
         theEdge = theEdge->next)
    {
        InterpolateEdge(
            theEdge->v[tail]->p,
            theEdge->e[tail][right]->v[tip] == theEdge->v[tail] ?
                theEdge->e[tail][right]->v[tail]->p :
                theEdge->e[tail][right]->v[tip]->p,
            aSpaceType,
            theEdge->beam_vertices[0]);
        // texture coordinates (u,v) = (0,0)

        InterpolateEdge(
            theEdge->v[tail]->p,
            theEdge->e[tail][left]->v[tip] == theEdge->v[tail] ?
                theEdge->e[tail][left]->v[tail]->p :
                theEdge->e[tail][left]->v[tip]->p,
            aSpaceType,
            theEdge->beam_vertices[1]);
        // texture coordinates (u,v) = (0,1)

        InterpolateEdge(
            theEdge->v[tip]->p,
            theEdge->e[tip][left]->v[tip] == theEdge->v[tip] ?
                theEdge->e[tip][left]->v[tail]->p :
                theEdge->e[tip][left]->v[tip]->p,
            aSpaceType,
            theEdge->beam_vertices[2]);
        // texture coordinates (u,v) = (1,1)

        InterpolateEdge(
            theEdge->v[tip]->p,
            theEdge->e[tip][right]->v[tip] == theEdge->v[tip] ?
                theEdge->e[tip][right]->v[tail]->p :
                theEdge->e[tip][right]->v[tip]->p,
            aSpaceType,
            theEdge->beam_vertices[3]);
        // texture coordinates (u,v) = (1,0)
    }
}


void CDirichletDomain::InterpolateEdge(
    double      aNearPoint[4],
    double      aFarPoint[4],
    SpaceTypeDD aSpaceType,
    GLfloat     aBeamVertex[3])
{
    double      u[4],
                v[4],
                theResult[4];
    double      theTheta;
    int         i;

    //  Construct an orthonormal basis {u,v}.
    for (i = 0; i < 4; i++)
        u[i] = aFarPoint[i] + aNearPoint[i];
    Normalize(u, aSpaceType);

    for (i = 0; i < 4; i++)
        v[i] = aFarPoint[i] - aNearPoint[i];
    Normalize(v, aSpaceType);

    //  Locate aNearPoint relative to {u,v},
    //  and let theResult be BEAM_WIDTH closer to the center.

    switch (aSpaceType)
    {
        case SpaceSphericalDD:
            theTheta = acos(InnerProduct(aNearPoint, u, aSpaceType));
            theTheta -= BEAM_WIDTH;
            for (i = 0; i < 4; i++)
                theResult[i] = cos(theTheta) * u[i]  -  sin(theTheta) * v[i];
            break;

        case SpaceFlatDD:
            for (i = 0; i < 4; i++)
                theResult[i] = aNearPoint[i]  +  BEAM_WIDTH * v[i];
            break;

        case SpaceHyperbolicDD:
            theTheta = ArcCosh(-InnerProduct(aNearPoint, u, aSpaceType));
            theTheta -= BEAM_WIDTH;
            for (i = 0; i < 4; i++)
                theResult[i] = cosh(theTheta) * u[i]  -  sinh(theTheta) * v[i];
            break;
    }

    for (i = 0; i < 3; i++)
        aBeamVertex[i] = (GLfloat)( theResult[i] / theResult[3] );
}

void CDirichletDomain::Normalize(
    double      aVector[4],
    SpaceTypeDD aSpaceType)
{
    double  theLengthSquared;
    int     i;

    theLengthSquared = InnerProduct(aVector, aVector, aSpaceType);
    assert(theLengthSquared != 0.0);
    for (i = 0; i < 4; i++)
        aVector[i] /= sqrt(fabs(theLengthSquared));
}

double CDirichletDomain::InnerProduct(
    double      u[4],
    double      v[4],
    SpaceTypeDD aSpaceType)
{
    switch (aSpaceType)
    {
        case SpaceSphericalDD:
            return u[0]*v[0] + u[1]*v[1] + u[2]*v[2] + u[3]*v[3];

        case SpaceFlatDD:
            //  In Euclidean space the inner product is a bit slippery.
            //  The following definition does the job.
            if (u[3]*v[3] != 0.0)
                return u[3]*v[3];
            else
                return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];

        case SpaceHyperbolicDD:
            return u[0]*v[0] + u[1]*v[1] + u[2]*v[2] - u[3]*v[3];

        default:    //  keep the compiler happy
            return 0.0;
    }
}

double CDirichletDomain::ArcCosh(double x)
{
    if (x >= 1.0)
        return log(x + sqrt(x*x - 1.0));
    else
        return 0.0;
}
