#include "MatrixAlgebra-GL.h"
#include <math.h>


void MatrixMultiply(
          Matrix    *ab,
    const Matrix    *a,
    const Matrix    *b)
{
    Matrix  temp;
    int     i,
            j,
            k;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
        {
            temp.m[i][j] = 0;
            for (k = 0; k < 4; k++)
                temp.m[i][j] += a->m[i][k] * b->m[k][j];
        }

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            ab->m[i][j] = temp.m[i][j];
}


void MatrixTranspose(
          Matrix    *aT,
    const Matrix    *a)
{
    Matrix  temp;
    int     i,
            j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            temp.m[i][j] = a->m[j][i];

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            aT->m[i][j] = temp.m[i][j];
}


void MatrixIdentity(
    Matrix  *a)
{
    int     i,
            j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            a->m[i][j] = (i == j) ? 1.0f : 0.0f;
}


void MatrixRotation(
    Matrix      *aMatrix,
    int         anAxis,
    double      anAngle)
{
    int i,
        j;

    //  Start with an identity matrix and then...
    MatrixIdentity(aMatrix);

    //  ...splice in the rotational part.
    i = (anAxis + 1)%3;
    j = (anAxis + 2)%3;
    aMatrix->m[i][i] = (float)  cos(anAngle);
    aMatrix->m[i][j] = (float)  sin(anAngle);
    aMatrix->m[j][i] = (float) -sin(anAngle);
    aMatrix->m[j][j] = (float)  cos(anAngle);
}


double MatrixDeterminant(Matrix *aMatrix)
{
    //  Here's a quick algorithm to compute the determinant.
    //  The determinant should always be +1 or -1,
    //  so precision isn't an issue.

    double  theDeterminant,
            theProduct;
    int     a,
            b,
            c,
            d,
            temp,
            parity,
            i,
            j,
            k;

    //  Initialize theDeterminant to zero.
    theDeterminant = 0.0;

    //  Set up the identity permutation.
    a = 0; b = 1; c = 2; d = 3;
    parity = 0;

    //  Cycle through all other permutations, keeping track of the parity.
    for (i = 0; i < 4; i++)
    {
        for (j = 0; j < 3; j++)
        {
            for (k = 0; k < 2; k++)
            {
                theProduct = aMatrix->m[0][a]
                           * aMatrix->m[1][b]
                           * aMatrix->m[2][c]
                           * aMatrix->m[3][d];

                if (parity == 0)
                    theDeterminant += theProduct;
                else
                    theDeterminant -= theProduct;

                //  swap the first two positions
                temp = b; b = a; a = temp;
                parity = ! parity;
            }

            //  cycle the first three positions
            temp = c; c = b; b = a; a = temp;
            //  parity doesn't change
        }

        //  cycle the whole permutation
        temp = d; d = c; c = b; b = a; a = temp;
        parity = ! parity;
    }

    return theDeterminant;
}


void MatrixInverse(
        Matrix  *anInverse,
  const Matrix  *aMatrix)
{
    double      row[4][8];
    double      *mm[4],
                *temp_row,
                multiple;
    int         i,
                j,
                k;

    for (i = 0; i < 4; i++)
        mm[i] = row[i];

    //  Copy m -- don't alter the original.
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            mm[i][j] = aMatrix->m[i][j];

    //  Initialize the four right hand columns to the identity.
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            mm[i][4 + j] = (i == j) ? 1.0 : 0.0;

    //  Forward elimination.
    for (j = 0; j < 4; j++)
    {
        //  Partial pivoting.
        for (i = j+1; i < 4; i++)
            if (fabs(mm[j][j]) < fabs(mm[i][j]))
            {
                temp_row    = mm[j];
                mm[j]       = mm[i];
                mm[i]       = temp_row;
            }

        //  Is the matrix singular?
//      assert(fabs(mm[j][j]) > SINGULAR_MATRIX_EPSILON);

        //  Divide through to get a 1.0 on the diagonal.
        multiple = 1.0 / mm[j][j];
        for (k = j; k < 8; k++)
            mm[j][k] *= multiple;

        //  Clear out that column.
        for (i = j+1; i < 4; i++)
        {
            multiple = mm[i][j];
            for (k = j; k < 8; k++)
                mm[i][k] -= multiple * mm[j][k];
        }
    }

    //  Back substitution.
    for (j = 4; --j >= 0; )
        for (i = j; --i >= 0; )
            for (k = 4; k < 8; k++)
                mm[i][k] -= mm[i][j] * mm[j][k];

    //  Copy out the solution.
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            anInverse->m[i][j] = (float) mm[i][4 + j];
}


BOOL MatrixIsIdentity(
    Matrix  *aMatrix)
{
    int i,
        j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            if (aMatrix->m[i][j] != (i == j ? 1.0 : 0.0))
                return FALSE;

    return TRUE;
}


BOOL MatrixIsApproximatelyIdentity(
    Matrix  *aMatrix,
    double  anEpsilon)
{
    int i,
        j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            if (fabs(aMatrix->m[i][j] - (i == j ? 1.0 : 0.0)) > anEpsilon)
                return FALSE;

    return TRUE;
}
