//  CurvedSpaces-Simple-D3D.cpp
//
//  This sample code accompanies the article
//
//      "Real-time rendering in curved spaces",
//      IEEE Computer Graphics and Applications, Nov-Dec 2002.
//
//  Interactive 3D animations in spherical, flat, and hyperbolic
//  spaces run in hardware on the user's 3D graphics card via Direct3D.
//
//  Different versions of this code serve different purposes.
//
//      CurvedSpaces-Simple-[D3D or GL].cpp
//
//          This stripped-down version of the program introduces
//          basic techniques of rendering in curved spaces.
//
//      CurvedSpaces-Complete-[D3D or GL].cpp
//
//          This more complete version of the program lets
//          the user explore a variety of 3-dimensional spaces,
//          and may be of interest to geometers and cosmologists
//          as well as computer graphics developers.
//
//  Readers may wish to run the richer CurvedSpaces-Complete.exe
//  to get a feel for what is possible, but read the source code
//  for the more modest CurvedSpaces-Simple.cpp to understand
//  the basics of the implementation.
//
//  Jeff Weeks
//  weeks@geometrygames.org
//  March 2004

//  This code requires the import libraries
//
//      d3dx.lib   ddraw.lib   user32.lib   gdi32.lib
//
//  For a debug build, link to d3dxd.lib instead of d3dx.lib.

#include "Macros-D3D.h"
#include "SceneryObjectCube-D3D.hpp"
#include "SceneryObjectOctahedron-D3D.hpp"
#include "CurvedSpacesResourceIDs-Simple.h"

//  Initial window size.
#define WINDOW_CLIENT_SIZE  400

//  What is the minimum acceptable size for the
//  window's client area?
#define MIN_CLIENT_SIZE 192

//  Pressing the up-arrow key increases the camera speed
//  by CAMERA_SPEED_INCREMENT.  Pressing the down-arrow key
//  decreases it by that amount.  Pressing the space bar
//  sets the camera speed to zero.
#define CAMERA_SPEED_INCREMENT  0.02f

//  After clicking into the scene, moving the mouse pitches and yaws
//  the camera at a rate proportional to CAMERA_ROTATION_RATE.
//  Holding the Ctrl key down while moving the mouse rolls the camera.
#define CAMERA_ROTATION_RATE    1.0f

//  The CLIPPING_DISTANCE should be as close to the camera
//  as possible without inducing z-buffer related artifacts.
//  In a scene of depth ~1, a clipping distance of 0.005f
//  or greater works fine with an ordinary 16-bit depth buffer.
//  [Exercise for advanced readers:  Why would w-buffering fail
//  in spherical rendering?]
#define CLIPPING_DISTANCE   0.01f

//  Background color.
#define BACKGROUND_COLOR    0xFFC0C0C0


//  Basic curved space rendering proceeds without knowing
//  the curvature of space.  The exceptions are in
//
//  (1) MatrixTranslation() which must know whether to create
//      a spherical, flat, or hyperbolic rotation matrix when
//      translating the camera or scenery position,
//
//  (2) NormalizeCameraPosition() which maintains an orthogonal
//      matrix for gCameraPosition, and
//
//  (3) Render() which in the spherical case must draw the back
//      hemisphere separately from the front hemisphere, as
//      explained in the accompanying article.
//
//  In CurvedSpaces-Simple.cpp these are the *only* uses of
//  the SpaceType.  In CurvedSpaces-Complete.cpp the SpaceType
//  is also taken into account when computing fog effects and
//  when constructing certain types of scenery.
typedef enum
{
    SpaceSpherical = 0,
    SpaceFlat,
    SpaceHyperbolic
} SpaceType;


//  The possible projection matrices are indexed according to whether
//  we are mapping the viewing banana into the front half of the
//  clipping box in projection space, into the back half, or into
//  the whole box. 
typedef enum
{
    BananaFull,
    BananaFront,
    BananaBack
} RangeType;



//  Is the current space spherical, flat, or hyperbolic?
static SpaceType        gSpaceType          = SpaceFlat;

//  Keep track of the camera's position and speed.
static D3DXMATRIX       gCameraPosition     = D3DXMATRIX(
                                                1,0,0,0,
                                                0,1,0,0,
                                                0,0,1,0,
                                                0,0,0,1);
static float            gCameraSpeed        = 0.0f;

//  This sample code uses only two scenery objects, but a more
//  interesting program -- say a game in a hypersphere -- would
//  use many.  Each may be a subclass of CSceneryObject.
#define NUM_SCENERY_OBJECTS 2
static CSceneryObject   *gSceneryObjects[NUM_SCENERY_OBJECTS]   = {NULL, NULL};

//  The accompanying article "Real-time rendering in curved spaces"
//  explains the banana projection matrix.  For most purposes
//  gProjectionMatrix[BananaFull] works fine, but if we want to see
//  the back half of spherical space, we'll need to use
//  gProjectionMatrix[BananaFront] and gProjectionMatrix[BananaBack]
//  instead.
static D3DXMATRIX   gProjectionMatrix[3] =
{
    //  The BananaFull matrix takes the viewing banana's corners
    //  to the corners of the clipping box in projection space.
    D3DXMATRIX( 1, 0,        0,                   0,
                0, 1,        0,                   0,
                0, 0,      0.50f,                 1,
                0, 0, -0.50f * CLIPPING_DISTANCE, 0),

    //  TheBananaFront matrix maps the viewing banana into
    //  the front half of the clipping box in projection space.
    D3DXMATRIX( 1, 0,        0,                   0,
                0, 1,        0,                   0,
                0, 0,      0.25f,                 1,
                0, 0, -0.25f * CLIPPING_DISTANCE, 0),

    //  The BananaBack matrix maps the viewing banana into
    //  the back half of the clipping box in projection space.
    //
    //  2003/02/17  Include a factor of -Id in the BananaBack
    //  matrix to invert the scenery, so we don't need to invert
    //  the camera at render time.
    D3DXMATRIX( -1,  0,        0,                   0,
                 0, -1,        0,                   0,
                 0,  0,     -0.75f,                -1,
                 0,  0, 0.25f * CLIPPING_DISTANCE,  0)
};

//  The program uses a single window.
static HWND gWindow = NULL;

//  The program maintains a D3DXContext, which is replaced
//  whenever the user changes the bit depth.
static ID3DXContext *gD3DXContext = NULL;

//  Should we display the frame rate?
static BOOL gShowFrameRate = FALSE;


static BOOL             InitSceneryObjects(void);
static void             SetSceneryPositions(void);
static void             BuildWindow(HINSTANCE hInstance);
static void             RestartGraphics(void);
static HRESULT          InitGraphics(void);
static void             ShutDownGraphics(void);
static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static void             MenuCommand(WORD aCommand, HWND aParentWindow);
static BOOL FAR PASCAL  TrivialDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static void             Update(void);
static void             NormalizeCameraPosition(void);
static void             Render(void);
static void             HandleWindowedModeChanges(ID3DXContext *aD3DXContext);
static void             MatrixTranslation(D3DXMATRIX *aMatrix, int anAxis, double aDistance);
static void             PrependTranslation(D3DXMATRIX *aMatrix, int anAxis, double aDistance);
static void             PrependRotation(D3DXMATRIX *aMatrix, float aYaw, float aPitch, float aRoll);
static void             InterpretError(HRESULT hr);


int PASCAL WinMain(
    HINSTANCE   hInstance,
    HINSTANCE   hPrevInstance, 
    LPSTR       lpszCmdLine,
    int         nCmdShow) 
{
    HRESULT hr;
    MSG     theMessage;
    int     i;

    //  Initialize the SceneryObjects.
    if (InitSceneryObjects() == FALSE)
        goto CleanUpAndExit;

    //  Initialize the window.
    BuildWindow(hInstance);

    //  Initialize the Direct3DX graphics.
    if (FAILED(hr = InitGraphics()))
    {
        InterpretError(hr);
        goto CleanUpAndExit;
    }

    //  Process messages until WM_QUIT is received.
    //  Render the scene during idle time.
    while (TRUE)
    {
        if (PeekMessage(&theMessage, NULL, 0, 0, PM_REMOVE) != 0)
        {
            TranslateMessage(&theMessage);
            DispatchMessage(&theMessage);

            if (theMessage.message == WM_QUIT)
                break;
        }
        else
        {
            if (IsIconic(gWindow) == FALSE)
            {
                Update();
                Render();

                //	When running under Windows ME
                //  this program is apparently "too fast"
                //  and leads to jerky mouse motion.
                //  Sleeping solves the problem and
                //  keeps the mouse action smooth.
                Sleep(1);
			}
            else
                Sleep(100);
        }
    }

CleanUpAndExit:

    //  Shutdown Direct3D.
    ShutDownGraphics();

    //  Release the window.
    if (gWindow != NULL)
        DestroyWindow(gWindow);

    //  Delete the SceneryObjects.
    for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
        SAFE_DELETE(gSceneryObjects[i])

    return 0;
}


static BOOL InitSceneryObjects(void)
{
    gSceneryObjects[0] = new CSceneryObjectCube();
    gSceneryObjects[1] = new CSceneryObjectOctahedron();

    if (gSceneryObjects[0] == NULL
     || gSceneryObjects[1] == NULL)
        return FALSE;

    SetSceneryPositions();

    return TRUE;
}


static void SetSceneryPositions(void)
{
    D3DXMATRIX  thePosition;

    MatrixTranslation(&thePosition, 2, 0.2);
    gSceneryObjects[0]->SetPosition(&thePosition);

    MatrixTranslation(&thePosition, 2, 0.4);
    gSceneryObjects[1]->SetPosition(&thePosition);
}


static void BuildWindow(
    HINSTANCE   hInstance)
{
    //  Register the window class.
    WNDCLASS wndClass =
    {
        CS_HREDRAW | CS_VREDRAW,
        WndProc,
        0,
        0,
        hInstance,
        LoadIcon(hInstance, "CurvedSpacesSimpleIcon"),
        LoadCursor(hInstance, "CurvedSpacesCursor"),
        (HBRUSH)GetStockObject(BLACK_BRUSH),
        TEXT("CurvedSpacesMenu"),
        TEXT("Curved Spaces Window Class")
    };
    RegisterClass(&wndClass);

    //  Create the window.
    gWindow = CreateWindow (
        TEXT("Curved Spaces Window Class"),
        TEXT("Sample Code (Direct3D)"),
        WS_OVERLAPPEDWINDOW | WS_VISIBLE,
        CW_USEDEFAULT, CW_USEDEFAULT,
        WINDOW_CLIENT_SIZE + 2 * GetSystemMetrics(SM_CXSIZEFRAME),
        WINDOW_CLIENT_SIZE + GetSystemMetrics(SM_CYCAPTION)
                           + GetSystemMetrics(SM_CYMENU)
                           + 2 * GetSystemMetrics(SM_CYSIZEFRAME),
        NULL, NULL, hInstance, NULL);
}



static void RestartGraphics(void)
{
    if (IsIconic(gWindow))
        return;

    ShutDownGraphics();

    if (FAILED(InitGraphics()))
    {
        MessageBox(NULL, "Can't reinitialize graphics.  Bye.", "We're cooked.", MB_OK);
        PostQuitMessage(0);
    }
}


static HRESULT InitGraphics(void)
{
    HRESULT hr;
    int     i;

    static DWORD    theHardwareSupportLevel[3] =
    {
        D3DX_HWLEVEL_TL,        //  hardware T&L and rasterization
        D3DX_HWLEVEL_RASTER,    //  hardware rasterization
        D3DX_HWLEVEL_2D         //  software rendering
    };

    //  Initialize the Direct3DX graphics.
    if (FAILED(hr = D3DXInitialize()))
        return hr;

    //  Create a Direct3DX Context.
    //
    //  Technical Note:  In theory we should be able to pass
    //  D3DX_DEFAULT for the deviceIndex and let D3DXCreateContext()
    //  pick the best available level of hardware support.
    //  In practice Direct3D sometimes shoots for a level of support
    //  that isn't available at the current bit depth, and then fails.
    //  For example, on my own graphics card it succeeds with
    //  16-bit color and fails with 32-bit color.
    //  To be safe, let's step through the support levels manually.
    i = 0;
    do
    {
        hr = D3DXCreateContext( theHardwareSupportLevel[i],
                                0,
                                gWindow,
                                D3DX_DEFAULT,
                                D3DX_DEFAULT,
                                &gD3DXContext);

    } while (FAILED(hr) && ++i < 3);

    return hr;
}


static void ShutDownGraphics(void)
{
    //  Release the D3DX Context.
    SAFE_RELEASE(gD3DXContext)

    //  Shut down D3DX.
    D3DXUninitialize();
}


//  Callback function to process window messages.
static LRESULT CALLBACK WndProc(
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    RECT            theRect;
    POINT           theNewCursorLocation;
    float           dx,
                    dy;
    int             i;
    static POINT    theWindowCenter = {0,0};
    static BOOL     theCursorIsHidden = FALSE;

    switch (uMsg)
    {
        case WM_CLOSE:
            PostQuitMessage(0);
            return 0;

        case WM_DESTROY:
            gWindow = NULL;
            return 0;

        case WM_KEYDOWN:
            switch (wParam) //  virtual key code
            {
                case VK_UP:
                    gCameraSpeed += CAMERA_SPEED_INCREMENT;
                    break;

                case VK_DOWN:
                    gCameraSpeed -= CAMERA_SPEED_INCREMENT;
                    break;

                case ' ':
                    gCameraSpeed = 0;
                    break;

                case VK_ESCAPE:
                    //  The user wants to stop flying.
                    //  ReleaseCapture() generates a WM_CAPTURECHANGED message
                    //  which will be handled below.
                    ReleaseCapture();
                    break;
            }
            return 0;

        case WM_COMMAND:
            MenuCommand(LOWORD(wParam), hWnd);
            return 0;

        case WM_LBUTTONDOWN:
            //  The user is clicking into the scene to steer.
            //  Capture the mouse and hide the cursor.
            SetCapture(hWnd);
            SetCursor(NULL);
            GetWindowRect(hWnd, &theRect);
            theWindowCenter.x = (theRect.right  + theRect.left)/2;
            theWindowCenter.y = (theRect.bottom + theRect.top )/2;
            SetCursorPos(theWindowCenter.x, theWindowCenter.y);
            theCursorIsHidden = TRUE;
            return 0;

        case WM_RBUTTONDOWN:
            //  The user wants to stop flying.
            //  ReleaseCapture() generates a WM_CAPTURECHANGED message
            //  which will be handled below.
            ReleaseCapture();
            return 0;

        case WM_CAPTURECHANGED:
            //  The WM_CAPTURECHANGED message typically results
            //  from our own call to ReleaseCapture(),
            //  but it could arise in other ways, for example,
            //  if some other program grabs the mouse.
            theCursorIsHidden = FALSE;
            return 0;

        case WM_MOUSEMOVE:
            //  When the cursor is hidden, mouse movements steer the camera.
            if (theCursorIsHidden)
            {
                GetCursorPos(&theNewCursorLocation);

                if (theNewCursorLocation.x != theWindowCenter.x
                 || theNewCursorLocation.y != theWindowCenter.y)
                {
                    GetWindowRect(hWnd, &theRect);

                    dx = (float)(theNewCursorLocation.x - theWindowCenter.x)
                       / (float)(theRect.right  - theRect.left);
                    dy = (float)(theNewCursorLocation.y - theWindowCenter.y)
                       / (float)(theRect.bottom - theRect.top );

                    SetCursorPos(theWindowCenter.x, theWindowCenter.y);

                    //  Allow full six-degrees-of-freedom navigation.
                    if (wParam & MK_SHIFT)
                    {
                        //  Translate
                        if (wParam & MK_CONTROL)
                            PrependTranslation(&gCameraPosition, 2, -dy);
                        else
                        {
                            PrependTranslation(&gCameraPosition, 0,  dx);
                            PrependTranslation(&gCameraPosition, 1, -dy);
                        }
                    }
                    else
                    {
                        //  Rotate
                        if (wParam & MK_CONTROL)   
                            PrependRotation(&gCameraPosition, 0, 0, -dx);                         
                        else
                            PrependRotation(&gCameraPosition, dx, dy, 0);
                    }
                }
            }
            return 0;

        case WM_SIZE:

            //  Tell Direct3D about the new window size...
            if (gD3DXContext != NULL)
                gD3DXContext->Resize(LOWORD(lParam), HIWORD(lParam));
 
            //  ...and adjust the projection matrix to accommodate it.
            for (i = 0; i < 3; i++)
                gProjectionMatrix[i].m[1][1] = (float)LOWORD(lParam) / (float)HIWORD(lParam);

            return 0;

        case WM_GETMINMAXINFO:
            // Prevent the client area from going smaller
            //  than MIN_CLIENT_SIZE.
            ((MINMAXINFO*)lParam)->ptMinTrackSize.x =
                MIN_CLIENT_SIZE
                + 2 * GetSystemMetrics(SM_CXSIZEFRAME);
            ((MINMAXINFO*)lParam)->ptMinTrackSize.y =
                MIN_CLIENT_SIZE
                + GetSystemMetrics(SM_CYCAPTION)
                + GetSystemMetrics(SM_CYMENU)
                + 2 * GetSystemMetrics(SM_CYSIZEFRAME);
            return 0;

        case WM_INITMENUPOPUP:

            CheckMenuItem(  (HMENU)wParam, ID_SPACE_SPHERICAL,
                gSpaceType == SpaceSpherical  ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_SPACE_FLAT,
                gSpaceType == SpaceFlat       ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_SPACE_HYPERBOLIC,
                gSpaceType == SpaceHyperbolic ? MF_CHECKED : MF_UNCHECKED);

            return 0;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
}


static void MenuCommand(
    WORD    aCommand,
    HWND    aParentWindow)
{
    switch (aCommand)
    {
        //  file menu

        case ID_FILE_EXIT:
            PostQuitMessage(0);
            break;

        //  space menu

        case ID_SPACE_SPHERICAL:
            gSpaceType = SpaceSpherical;
            D3DXMatrixIdentity(&gCameraPosition);
            gCameraSpeed = 0;
            SetSceneryPositions();
            break;

        case ID_SPACE_FLAT:
            gSpaceType = SpaceFlat;
            D3DXMatrixIdentity(&gCameraPosition);
            gCameraSpeed = 0;
            SetSceneryPositions();
            break;

        case ID_SPACE_HYPERBOLIC:
            gSpaceType = SpaceHyperbolic;
            D3DXMatrixIdentity(&gCameraPosition);
            gCameraSpeed = 0;
            SetSceneryPositions();
            break;

        //  help menu

        case ID_HELP_INSTRUCTIONS:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("INSTRUCTIONSDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;

        case ID_HELP_ABOUT:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("ABOUTDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;
    }
}

static BOOL FAR PASCAL TrivialDialogProc(
    HWND    hDlg,
    UINT    msg,
    WPARAM  wparam,
    LPARAM  lParam)
{
    switch (msg)
    {
        case WM_INITDIALOG:
            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wparam))
            {
                case IDOK:
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                default:
                    return FALSE;
            }

        default:
            return FALSE;
    }
}


static void Update(void)
{
    //  Tell the camera and each scenery object how many
    //  seconds have elapsed since the last Update().

    static unsigned int theStartTime    = 0,
                        theNumFrames    = 0;
    static double       theFrameRate    = 30.0;
    static BOOL         theFirstSecond  = TRUE;
    
    unsigned int    theCurrentTime,
                    theElapsedTime;
    double          theNewFrameRate;
    char            theWindowTitle[128];
    int             i;

    //  Get the time in milliseconds.
    theCurrentTime = GetTickCount();

    //  One-time initialization.
    if (theStartTime == 0)
    {
        theStartTime = theCurrentTime;
    }
    else
    {
        //  Count the number of frames drawn in approximately one second.
        theNumFrames++;
        theElapsedTime = theCurrentTime - theStartTime;
        if (theFirstSecond == TRUE && theElapsedTime > 0)
        {
            //  During the first second of operation,
            //  revise theFrameRate after each clock tick.
            //  At first it won't be super accurate, but at least
            //  it'll be better than our arbitrary initial guess.
            theFrameRate = theNumFrames / (0.001 * theElapsedTime);
        }
        if (theElapsedTime > 1000)
        {
            theFirstSecond = FALSE;

            theNewFrameRate = theNumFrames / (0.001 * theElapsedTime);
            theNumFrames = 0;
            theStartTime = theCurrentTime;
        
            //  If the user holds down a menu, our program gets suspended
            //  but the clock keeps running.  As a result, the nominal
            //  frame rate is too low and the scenery moves too fast
            //  in compensation.  To avoid this problem, don't let
            //  the nominal frame rate decrease too fast.
            if (theNewFrameRate < 0.8 * theFrameRate)
                theNewFrameRate = 0.8 * theFrameRate;
        
            theFrameRate = theNewFrameRate;

            if (gShowFrameRate)
            {
                wsprintf(theWindowTitle, "%3d frames per second", (int)theFrameRate);
                SetWindowText(gWindow, theWindowTitle);
            }
        }
    }

    //  Update the camera position.
    //  That is, move the camera a tiny bit forward.
    PrependTranslation(&gCameraPosition, 2, gCameraSpeed / theFrameRate);

    //  Normalize gCameraPosition to keep it legal for the current SpaceType.
    NormalizeCameraPosition();

    //  Update the SceneryObjects.
    for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
        gSceneryObjects[i]->Update(1.0 / theFrameRate);
}


static void NormalizeCameraPosition(void)
{
    //  Numerical errors can accumulate and force the camera position
    //  matrix "out of round", in the sense that its rows are no
    //  longer orthonormal.  This effect is small in spherical and
    //  flat spaces, but can be significant in hyperbolic spaces,
    //  especially if the camera travels far from the origin.

    //  The "Gram-Schmidt process" consists of rescaling each
    //  row to restore unit length, and subtracting small multiples
    //  of one row from another to restore orthogonality.
    //  Here we carry out a first-order approximation to the
    //  Gram-Schmidt process.  That is, we normalize each row
    //  to unit length, but then assume that the subsequent
    //  orthogonalization step doesn't spoil the unit length.
    //  This assumption will be well satisfied because small
    //  first order changes orthogonal to a given vector affect
    //  its length only to second order.

    //  If your primary goal is to understand the techniques
    //  of curved space rendering, you may safely skip over
    //  the details of this orthonormalization. 

    D3DXMATRIX  m1,
                m2,
                theProduct;
    float       theFactor;
    int         i,
                j,
                k;

    //  Concoct a matrix giving the inner products
    //  of gCameraPosition's rows with one another.
    //  (This is a bit ad hoc, but it works.)

    m1 = gCameraPosition;
    D3DXMatrixTranspose(&m2, &m1);

    switch (gSpaceType)
    {
        case SpaceSpherical:
            //  no change required
            break;

        case SpaceFlat:

            //  zero out the mixed terms in both m1 and m2

            m1.m[3][0] = 0.0f;
            m1.m[3][1] = 0.0f;
            m1.m[3][2] = 0.0f;

            m2.m[0][3] = 0.0f;
            m2.m[1][3] = 0.0f;
            m2.m[2][3] = 0.0f;

            break;

        case SpaceHyperbolic:

            //  negate the mixed terms in m2 only

            m2.m[0][3] = - m2.m[0][3];
            m2.m[1][3] = - m2.m[1][3];
            m2.m[2][3] = - m2.m[2][3];

            m2.m[3][0] = - m2.m[3][0];
            m2.m[3][1] = - m2.m[3][1];
            m2.m[3][2] = - m2.m[3][2];

            break;
    }

    D3DXMatrixMultiply(&theProduct, &m1, &m2);

    //  Normalize the length of each row of gCameraPosition.
    for (i = 0; i < 4; i++)
    {
        if (theProduct.m[i][i] <= 0.0f)
            exit(1);

        theFactor = (float) (1.0 / sqrt(theProduct.m[i][i]));

        for (j = 0; j < 4; j++)
            gCameraPosition.m[i][j] *= theFactor;
    }

    //  Make the rows of gCameraPosition mutually orthogonal.
    for (i = 4; --i >= 0; )         // iterate backwards
        for (j = i; --j >= 0; )     // iterate backwards
        {
            theFactor = theProduct.m[j][i];

            for (k = 0; k < 4; k++)
                gCameraPosition.m[j][k] -=
                    theFactor * gCameraPosition.m[i][k];
                    
        }
}


static void Render(void)
{
    LPDIRECT3DDEVICE7   theD3DDevice;
    D3DXMATRIX          theViewMatrix;
    LPDIRECTDRAW7       theDD = NULL;
    int                 i;

    if (gWindow == NULL
     || IsIconic(gWindow))
        return;

    try
    {
        theD3DDevice = gD3DXContext->GetD3DDevice();
        if (theD3DDevice == NULL)
            throw new CException("Couldn't get D3DDevice from D3DContext.");

        //  Using w buffering instead of z buffering would give incorrect
        //  results in the spherical case, so let's explicitly set
        //  D3DRENDERSTATE_ZENABLE to D3DZB_TRUE (as opposed to
        //  D3DZBUFFERTYPE's third possibility, which is D3DZB_USEW).
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_ZENABLE, D3DZB_TRUE))

        //  For this simple example, let's show both sides of all surfaces.
        //  In more complex software, where the frame rate is an issue,
        //  set this to the default D3DCULL_CCW (or omit this line entirely).
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_NONE))

        //  Let's keep things simple and give all objects a constant illumination.
        //  (Direct3D 8.0 might support custom spherical and hyperbolic lighting,
        //  but I haven't yet investigated that possibility.)
        D3DTRY(theD3DDevice->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE))

        D3DTRY(theD3DDevice->BeginScene())

        //  Initialize the render target and z-buffer.
        D3DTRY(theD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, BACKGROUND_COLOR, 1.0f, 0))

        //  The view matrix (which maps world space into camera space) is the
        //  inverse of gCameraPosition (which maps camera space into world space).
        D3DXMatrixInverse(&theViewMatrix, NULL, &gCameraPosition);
        D3DTRY(theD3DDevice->SetTransform(
            D3DTRANSFORMSTATE_VIEW,
            &D3DMATRIX(theViewMatrix)))

        if (gSpaceType != SpaceSpherical)
        {
            //  In flat and hyperbolic spaces standard rendering
            //  works great.  In spherical space it would show
            //  only objects in the front hemisphere.

            //  Set the BananaFull projection matrix.
            //  (A traditional projection matrix would also work fine here.)
            D3DTRY(theD3DDevice->SetTransform(
                D3DTRANSFORMSTATE_PROJECTION,
                &D3DMATRIX(gProjectionMatrix[BananaFull])))

            //  Render the scenery.
            for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
                gSceneryObjects[i]->Render(theD3DDevice);
        }
        else
        {
            //  In spherical space we render the front and back
            //  hemispheres separately. The accompanying article
            //  "Real-time rendering in curved spaces" explains why.

            //  Set the BananaFront projection matrix, which maps
            //  the front hemisphere viewing banana into the front half
            //  of the clipping box (0 <= z <= 1/2).
            D3DTRY(theD3DDevice->SetTransform(
                D3DTRANSFORMSTATE_PROJECTION,
                &D3DMATRIX(gProjectionMatrix[BananaFront])))

            //  Render the scenery.
            for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
                gSceneryObjects[i]->Render(theD3DDevice);

            //  Set the BananaBack projection matrix, which maps
            //  the back hemisphere viewing banana into the back half
            //  of the clipping box (1/2 <= z <= 1).
            D3DTRY(theD3DDevice->SetTransform(
                D3DTRANSFORMSTATE_PROJECTION,
                &D3DMATRIX(gProjectionMatrix[BananaBack])))

            //  Render the scenery.
            for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
                gSceneryObjects[i]->Render(theD3DDevice);
        }

        D3DTRY(theD3DDevice->EndScene())

        SAFE_RELEASE(theD3DDevice)

        D3DTRY(gD3DXContext->DrawDebugText(0.1f, 0.1f, 0x00FFFF, "Choose Instructions... from the Help menu."))

        theDD = gD3DXContext->GetDD();
        if (theDD != NULL)
            SAFE_RELEASE(theDD);

        D3DTRY(gD3DXContext->UpdateFrame(0))
    }
    catch (CException *anException)
    {
        SAFE_DELETE(anException)
        if (theD3DDevice != NULL)
            theD3DDevice->EndScene();
        SAFE_RELEASE(theD3DDevice)

        //  The most likely reason we ended up here is that
        //  we lost our surfaces while the computer was on Stand By,
        //  or when the monitor settings were changed.
        HandleWindowedModeChanges(gD3DXContext);
    }
    catch (...)
    {
        MessageBox(NULL, "Unknown exception -- couldn't render scene.", "Rendering Error", MB_OK);
        if (theD3DDevice != NULL)
            theD3DDevice->EndScene();
        SAFE_RELEASE(theD3DDevice)
    }
}


static void HandleWindowedModeChanges(ID3DXContext *aD3DXContext)
{
    //  Lots of things can go wrong when different programs
    //  share a graphics card.

    LPDIRECTDRAW7   theDD;

    theDD = aD3DXContext->GetDD();

    //  If some other program grabbed the video card
    //  in exclusive mode, sit tight until it's done.
    while (theDD->TestCooperativeLevel() == DDERR_EXCLUSIVEMODEALREADYSET)
        Sleep(500);

    SAFE_RELEASE(theDD)

    //  Rather than trying to figure out the most efficient
    //  solution for each of the many possible cases,
    //  let's just trash our D3DX Context and start over.
    RestartGraphics();
}


static void MatrixTranslation(
    D3DXMATRIX  *aMatrix,
    int         anAxis,
    double      aDistance)
{

    D3DXMatrixIdentity(aMatrix);

    switch (gSpaceType)
    {
        case SpaceSpherical:
            aMatrix->m[anAxis][anAxis]    = (float)  cos(aDistance);
            aMatrix->m[3][3]              = (float)  cos(aDistance);
            aMatrix->m[anAxis][3]         = (float) -sin(aDistance);
            aMatrix->m[3][anAxis]         = (float)  sin(aDistance);
            break;

        case SpaceFlat:
            aMatrix->m[3][anAxis]         = (float) aDistance;
            break;

        case SpaceHyperbolic:
            aMatrix->m[anAxis][anAxis]    = (float) cosh(aDistance);
            aMatrix->m[3][3]              = (float) cosh(aDistance);
            aMatrix->m[anAxis][3]         = (float) sinh(aDistance);
            aMatrix->m[3][anAxis]         = (float) sinh(aDistance);
            break;
    }
}


static void PrependTranslation(
    D3DXMATRIX  *aMatrix,
    int         anAxis,
    double      aDistance)
{
    D3DXMATRIX  theTranslation;

    MatrixTranslation(&theTranslation, anAxis, aDistance);
    D3DXMatrixMultiply( aMatrix,
                        &theTranslation,
                        aMatrix);
}


static void PrependRotation(
    D3DXMATRIX  *aMatrix,
    float       aYaw,
    float       aPitch,
    float       aRoll)
{
    D3DXMATRIX  temp;

    D3DXMatrixMultiply( aMatrix,
                        D3DXMatrixRotationYawPitchRoll(
                            &temp,
                            CAMERA_ROTATION_RATE * aYaw,
                            CAMERA_ROTATION_RATE * aPitch,
                            CAMERA_ROTATION_RATE * aRoll),
                        aMatrix);
}


static void InterpretError(
    HRESULT hr)
{
    char    theErrorString[256];

    D3DXGetErrorString(hr, 256, theErrorString);
    MessageBox(NULL, theErrorString, "D3DX Error", MB_OK);
}


void CheckD3DResult(HRESULT hr)
{
    if FAILED(hr)
    {
        char    *theMessage;

        theMessage = new char[256];
        D3DXGetErrorString(hr, 256, theMessage);
        throw new CException(theMessage, TRUE);
    }
}
