//  CurvedSpaces-Simple-GL.cpp
//
//  This sample code accompanies the article
//
//      "Real-time rendering in curved spaces",
//      IEEE Computer Graphics and Applications, Nov-Dec 2002.
//
//  Interactive 3D animations in spherical, flat, and hyperbolic
//  spaces run in hardware on the user's 3D graphics card via OpenGL.
//
//  Different versions of this code serve different purposes.
//
//      CurvedSpaces-Simple-[D3D or GL].cpp
//
//          This stripped-down version of the program introduces
//          basic techniques of rendering in curved spaces.
//
//      CurvedSpaces-Complete-[D3D or GL].cpp
//
//          This more complete version of the program lets
//          the user explore a variety of 3-dimensional spaces,
//          and may be of interest to geometers and cosmologists
//          as well as computer graphics developers.
//
//  Readers may wish to run the richer CurvedSpaces-Complete.exe
//  to get a feel for what is possible, but read the source code
//  for the more modest CurvedSpaces-Simple.cpp to understand
//  the basics of the implementation.
//
//  Jeff Weeks
//  weeks@geometrygames.org
//  March 2004

//  This code requires the import libraries
//
//      opengl32.lib   user32.lib   gdi32.lib

#include "Macros-GL.h"
#include "SceneryObjectCube-GL.hpp"
#include "SceneryObjectOctahedron-GL.hpp"
#include "CurvedSpacesResourceIDs-Simple.h"
#include <math.h>


//  Initial window size.
#define WINDOW_CLIENT_SIZE  400

//  What is the minimum acceptable size for the
//  window's client area?
#define MIN_CLIENT_SIZE 192

//  Pressing the up-arrow key increases the camera speed
//  by CAMERA_SPEED_INCREMENT.  Pressing the down-arrow key
//  decreases it by that amount.  Pressing the space bar
//  sets the camera speed to zero.
#define CAMERA_SPEED_INCREMENT  0.02f

//  After clicking into the scene, moving the mouse pitches and yaws
//  the camera at a rate proportional to CAMERA_ROTATION_RATE.
//  Holding the Ctrl key down while moving the mouse rolls the camera.
#define CAMERA_ROTATION_RATE    1.0f

//  The CLIPPING_DISTANCE should be as close to the camera
//  as possible without inducing z-buffer related artifacts.
//  In a scene of depth ~1, a clipping distance of 0.005f
//  or greater works fine with an ordinary 16-bit depth buffer.
//  [Exercise for advanced readers:  Why would w-buffering fail
//  in spherical rendering?]
#define CLIPPING_DISTANCE   0.01f

//  Background color.
#define BACKGROUND_COLOR    (0.75f, 0.75f, 0.75f, 0.00f)


//  Basic curved space rendering proceeds without knowing
//  the curvature of space.  The exceptions are in
//
//  (1) MatrixTranslation() which must know whether to create
//      a spherical, flat, or hyperbolic rotation matrix when
//      translating the camera or scenery position,
//
//  (2) NormalizeViewMatrix() which maintains an orthogonal
//      matrix for gViewMatrix, and
//
//  (3) Render() which in the spherical case must draw the back
//      hemisphere separately from the front hemisphere, as
//      explained in the accompanying article.
//
//  In CurvedSpaces-Simple.cpp these are the *only* uses of
//  the SpaceType.  In CurvedSpaces-Complete.cpp the SpaceType
//  is also taken into account when computing fog effects and
//  when constructing certain types of scenery.
typedef enum
{
    SpaceSpherical = 0,
    SpaceFlat,
    SpaceHyperbolic
} SpaceType;


//  The possible projection matrices are indexed according to whether
//  we are mapping the viewing banana into the front half of the
//  clipping box in projection space, into the back half, or into
//  the whole box. 
typedef enum
{
    BananaFull = 0,
    BananaFront,
    BananaBack
} RangeType;


//  Is the current space spherical, flat, or hyperbolic?
static SpaceType        gSpaceType          = SpaceFlat;

//  A matrix gCameraPosition would keep track of the camera's
//  position in world space.  Its inverse, gViewMatrix, keeps
//  track of the world's position in camera space.
static Matrix           gViewMatrix         = {{
                                                {1,0,0,0},
                                                {0,1,0,0},
                                                {0,0,1,0},
                                                {0,0,0,1}
                                              }};
static float            gCameraSpeed        = 0.0f;

//  This sample code uses only two scenery objects, but a more
//  interesting program -- say a game in a hypersphere -- would
//  use many.  Each may be a subclass of CSceneryObject.
#define NUM_SCENERY_OBJECTS 2
static CSceneryObject   *gSceneryObjects[NUM_SCENERY_OBJECTS]   = {NULL, NULL};

//  The accompanying article "Real-time rendering in curved spaces"
//  explains the banana projection matrices.  For most purposes
//  gProjectionMatrix[BananaFull] works fine, but if we want to see
//  the back half of spherical space, we'll need to use
//  gProjectionMatrix[BananaFront] and gProjectionMatrix[BananaBack]
//  instead.
static Matrix           gProjectionMatrix[3] =
    {
        //  The BananaFull matrix takes the viewing banana's corners
        //  to the corners of the clipping box in projection space.
        {{
            {1, 0,          0,           0},
            {0, 1,          0,           0},
            {0, 0,          0,          -1},
            {0, 0, -CLIPPING_DISTANCE,   0}
        }},

        //  TheBananaFront matrix maps the viewing banana into
        //  the front half of the clipping box in projection space.
        {{
            {1, 0,          0,           0},
            {0, 1,          0,           0},
            {0, 0,         0.5,         -1},
            {0, 0, -CLIPPING_DISTANCE/2, 0}
        }},

        //  The BananaBack matrix maps the viewing banana into
        //  the back half of the clipping box in projection space.
        //
        //  2003/02/17  Include a factor of -Id in the BananaBack
        //  matrix to invert the scenery, so we don't need to invert
        //  the camera at render time.
        {{
            {-1,  0,         0,           0},
            { 0, -1,         0,           0},
            { 0,  0,        0.5,          1},
            { 0,  0, CLIPPING_DISTANCE/2, 0}
        }}
    };

//  Should we display the frame rate?
static BOOL     gShowFrameRate      = FALSE;

//  The program uses a single window.
static HWND     gWindow             = NULL;
static HDC      gDeviceContext      = NULL;
static HGLRC    gRenderingContext   = NULL;


static BOOL             InitSceneryObjects(void);
static void             SetSceneryPositions(void);
static void             BuildWindow(HINSTANCE hInstance);
static BOOL             InitGraphics(void);
static void             ShutDownGraphics(void);
static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static void             MenuCommand(WORD aCommand, HWND aParentWindow);
static BOOL FAR PASCAL  TrivialDialogProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lParam);
static void             Update(void);
static void             NormalizeViewMatrix(void);
static void             Render(void);
static void             MatrixMultiply(Matrix *ab, const Matrix *a, const Matrix *b);
static void             MatrixTranspose(Matrix *aT, const Matrix *a);
static void             MatrixIdentity(Matrix *a);
static void             MatrixTranslation(Matrix *aMatrix, int anAxis, double aDistance);
static void             AppendTranslation(Matrix *aMatrix, int anAxis, double aDistance);
static void             MatrixRotation(Matrix *aMatrix, int anAxis, double anAngle);
static void             AppendRotation(Matrix *aMatrix, float aPitch, float aYaw, float aRoll);


int PASCAL WinMain(
    HINSTANCE   hInstance,
    HINSTANCE   hPrevInstance, 
    LPSTR       lpszCmdLine,
    int         nCmdShow) 
{
    MSG     theMessage;
    int     i;

    //  Initialize the SceneryObjects.
    if (InitSceneryObjects() == FALSE)
        goto CleanUpAndExit;

    //  Initialize the window.
    BuildWindow(hInstance);

    //  Initialize the OpenGL graphics.
    if (InitGraphics() == FALSE)
        goto CleanUpAndExit;

    //  Process messages until WM_QUIT is received.
    //  Render the scene during idle time.
    while (TRUE)
    {
        if (PeekMessage(&theMessage, NULL, 0, 0, PM_REMOVE) != 0)
        {
            TranslateMessage(&theMessage);
            DispatchMessage(&theMessage);

            if (theMessage.message == WM_QUIT)
                break;
        }
        else
        {
            if (IsIconic(gWindow) == FALSE)
            {
                Update();
                Render();
			}
            else
                Sleep(100);
        }
    }

CleanUpAndExit:

    //  Shut down OpenGL.
    ShutDownGraphics();

    //  Release the window.
    if (gWindow != NULL)
        DestroyWindow(gWindow);

    //  Delete the SceneryObjects.
    for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
        SAFE_DELETE(gSceneryObjects[i])

    return 0;
}


static BOOL InitSceneryObjects(void)
{
    gSceneryObjects[0] = new CSceneryObjectCube();
    gSceneryObjects[1] = new CSceneryObjectOctahedron();

    if (gSceneryObjects[0] == NULL
     || gSceneryObjects[1] == NULL)
        return FALSE;

    SetSceneryPositions();

    return TRUE;
}


static void SetSceneryPositions(void)
{
    Matrix  thePosition;

    MatrixTranslation(&thePosition, 2, -0.2);
    gSceneryObjects[0]->SetPosition(&thePosition);

    MatrixTranslation(&thePosition, 2, -0.4);
    gSceneryObjects[1]->SetPosition(&thePosition);
}


static void BuildWindow(
    HINSTANCE   hInstance)
{
    //  Register the window class.
    WNDCLASS wndClass =
    {
        CS_HREDRAW | CS_VREDRAW | CS_OWNDC,
        WndProc,
        0,
        0,
        hInstance,
        LoadIcon(hInstance, "CurvedSpacesSimpleIcon"),
        LoadCursor(hInstance, "CurvedSpacesCursor"),
        (HBRUSH)GetStockObject(BLACK_BRUSH),
        TEXT("CurvedSpacesMenu"),
        TEXT("Curved Spaces Window Class")
    };
    RegisterClass(&wndClass);

    //  Create the window.
    gWindow = CreateWindow (
        TEXT("Curved Spaces Window Class"),
        TEXT("Sample Code (OpenGL)"),
        WS_OVERLAPPEDWINDOW | WS_VISIBLE,
        CW_USEDEFAULT, CW_USEDEFAULT,
        WINDOW_CLIENT_SIZE + 2 * GetSystemMetrics(SM_CXSIZEFRAME),
        WINDOW_CLIENT_SIZE + GetSystemMetrics(SM_CYCAPTION)
                           + GetSystemMetrics(SM_CYMENU)
                           + 2 * GetSystemMetrics(SM_CYSIZEFRAME),
        NULL, NULL, hInstance, NULL);
}


static BOOL InitGraphics(void)
{
    GLuint  thePixelFormat;

    static PIXELFORMATDESCRIPTOR    thePixelFormatDescriptor =
    {
        sizeof(PIXELFORMATDESCRIPTOR),  //  size
        1,                              //  version number
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER, //  format
        PFD_TYPE_RGBA,                  //  preferred format is rgba (not palette)
        24,                             //  preferred color depth
        0, 0, 0, 0, 0, 0,               //  color bits ignored
        8,                              //  alpha buffer depth
        0,                              //  shift bit ignored
        0,                              //  no accumulation buffer
        0, 0, 0, 0,                     //  accumulation bits ignored
        16,                             //  16-bit Z-buffer  
        0,                              //  no stencil buffer
        0,                              //  no auxiliary buffer
        0,                              //  ignored
        0,                              //  underlay/overlay planes
        0,                              //  ignored
        0,                              //  underlay color
        0                               //  ignored
    };

    gDeviceContext = GetDC(gWindow);
    if (gDeviceContext == NULL)
    {
        MessageBox(NULL, "Can't get a device context to initialize OpenGL.", "GetDC() failed", MB_OK);
        return FALSE;
    }

    thePixelFormat = ChoosePixelFormat(gDeviceContext, &thePixelFormatDescriptor);
    if (thePixelFormat == 0)
    {
        MessageBox(NULL, "Can't get a pixel format to initialize OpenGL.", "ChoosePixelFormat() failed", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat(gDeviceContext, thePixelFormat, &thePixelFormatDescriptor) == FALSE)
    {
        MessageBox(NULL, "Can't set the pixel format to initialize OpenGL.", "SetPixelFormat() failed", MB_OK);
        return FALSE;
    }

    gRenderingContext = wglCreateContext(gDeviceContext);
    if (gRenderingContext == NULL)
    {
        MessageBox(NULL, "Can't get an OpenGL rendering context.", "wglCreateContext() failed", MB_OK);
        return FALSE;
    }

    if (wglMakeCurrent(gDeviceContext, gRenderingContext) == FALSE)
    {
        MessageBox(NULL, "Can't set the current rendering context.", "wglMakeCurrent() failed", MB_OK);
        return FALSE;
    }

    return TRUE;
}


static void ShutDownGraphics(void)
{
    wglMakeCurrent(NULL, NULL);

    if (gRenderingContext != NULL)
    {
        wglDeleteContext(gRenderingContext);
        gRenderingContext = NULL;
    }

    if (gDeviceContext != NULL)
    {
        ReleaseDC(gWindow, gDeviceContext);
        gDeviceContext = NULL;
    }
}


//  Callback function to process window messages.
static LRESULT CALLBACK WndProc(
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    RECT            theRect;
    POINT           theNewCursorLocation;
    float           dx,
                    dy;
    int             i;
    static POINT    theWindowCenter = {0,0};
    static BOOL     theCursorIsHidden = FALSE;

    switch (uMsg)
    {
        case WM_CLOSE:
            PostQuitMessage(0);
            return 0;

        case WM_DESTROY:
            gWindow = NULL;
            return 0;

        case WM_KEYDOWN:
            switch (wParam) //  virtual key code
            {
                case VK_UP:
                    gCameraSpeed += CAMERA_SPEED_INCREMENT;
                    break;

                case VK_DOWN:
                    gCameraSpeed -= CAMERA_SPEED_INCREMENT;
                    break;

                case ' ':
                    gCameraSpeed = 0;
                    break;

                case VK_ESCAPE:
                    //  The user wants to stop flying.
                    //  ReleaseCapture() generates a WM_CAPTURECHANGED message
                    //  which will be handled below.
                    ReleaseCapture();
                    break;
            }
            return 0;

        case WM_COMMAND:
            MenuCommand(LOWORD(wParam), hWnd);
            return 0;

        case WM_LBUTTONDOWN:
            //  The user is clicking into the scene to steer.
            //  Capture the mouse and hide the cursor.
            SetCapture(hWnd);
            SetCursor(NULL);
            GetWindowRect(hWnd, &theRect);
            theWindowCenter.x = (theRect.right  + theRect.left)/2;
            theWindowCenter.y = (theRect.bottom + theRect.top )/2;
            SetCursorPos(theWindowCenter.x, theWindowCenter.y);
            theCursorIsHidden = TRUE;
            return 0;

        case WM_RBUTTONDOWN:
            //  The user wants to stop flying.
            //  ReleaseCapture() generates a WM_CAPTURECHANGED message
            //  which will be handled below.
            ReleaseCapture();
            return 0;

        case WM_CAPTURECHANGED:
            //  The WM_CAPTURECHANGED message typically results
            //  from our own call to ReleaseCapture(),
            //  but it could arise in other ways, for example,
            //  if some other program grabs the mouse.
            theCursorIsHidden = FALSE;
            return 0;

        case WM_MOUSEMOVE:
            //  When the cursor is hidden, mouse movements steer the camera.
            if (theCursorIsHidden)
            {
                GetCursorPos(&theNewCursorLocation);

                if (theNewCursorLocation.x != theWindowCenter.x
                 || theNewCursorLocation.y != theWindowCenter.y)
                {
                    GetWindowRect(hWnd, &theRect);

                    dx = (float)(theNewCursorLocation.x - theWindowCenter.x)
                       / (float)(theRect.right  - theRect.left);
                    dy = (float)(theNewCursorLocation.y - theWindowCenter.y)
                       / (float)(theRect.bottom - theRect.top );

                    SetCursorPos(theWindowCenter.x, theWindowCenter.y);

                    //  Allow full six-degrees-of-freedom navigation.
                    if (wParam & MK_SHIFT)
                    {
                        //  Translate
                        if (wParam & MK_CONTROL)
                            AppendTranslation(&gViewMatrix, 2, -dy);
                        else
                        {
                            AppendTranslation(&gViewMatrix, 0, -dx);
                            AppendTranslation(&gViewMatrix, 1,  dy);
                        }
                    }
                    else
                    {
                        //  Rotate
                        if (wParam & MK_CONTROL)   
                            AppendRotation(&gViewMatrix,  0,  0, dx);                         
                        else
                            AppendRotation(&gViewMatrix, dy, dx,  0);
                    }
                }
            }
            return 0;

        case WM_SIZE:

            //  Tell OpenGL about the new window size...
            if (gWindow != NULL)
                glViewport(0, 0, LOWORD(lParam), HIWORD(lParam));
 
            //  ...and adjust the projection matrix to accommodate it.
            for (i = 0; i < 3; i++)
                gProjectionMatrix[i].m[1][1] = (float)LOWORD(lParam) / (float)HIWORD(lParam);

            return 0;

        case WM_GETMINMAXINFO:
            // Prevent the client area from going smaller
            //  than MIN_CLIENT_SIZE.
            ((MINMAXINFO*)lParam)->ptMinTrackSize.x =
                MIN_CLIENT_SIZE
                + 2 * GetSystemMetrics(SM_CXSIZEFRAME);
            ((MINMAXINFO*)lParam)->ptMinTrackSize.y =
                MIN_CLIENT_SIZE
                + GetSystemMetrics(SM_CYCAPTION)
                + GetSystemMetrics(SM_CYMENU)
                + 2 * GetSystemMetrics(SM_CYSIZEFRAME);
            return 0;

        case WM_INITMENUPOPUP:

            CheckMenuItem(  (HMENU)wParam, ID_SPACE_SPHERICAL,
                gSpaceType == SpaceSpherical  ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_SPACE_FLAT,
                gSpaceType == SpaceFlat       ? MF_CHECKED : MF_UNCHECKED);
            CheckMenuItem(  (HMENU)wParam, ID_SPACE_HYPERBOLIC,
                gSpaceType == SpaceHyperbolic ? MF_CHECKED : MF_UNCHECKED);

            return 0;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
}


static void MenuCommand(
    WORD    aCommand,
    HWND    aParentWindow)
{
    switch (aCommand)
    {
        //  file menu

        case ID_FILE_EXIT:
            PostQuitMessage(0);
            break;

        //  space menu

        case ID_SPACE_SPHERICAL:
            gSpaceType = SpaceSpherical;
            MatrixIdentity(&gViewMatrix);
            gCameraSpeed = 0;
            SetSceneryPositions();
            break;

        case ID_SPACE_FLAT:
            gSpaceType = SpaceFlat;
            MatrixIdentity(&gViewMatrix);
            gCameraSpeed = 0;
            SetSceneryPositions();
            break;

        case ID_SPACE_HYPERBOLIC:
            gSpaceType = SpaceHyperbolic;
            MatrixIdentity(&gViewMatrix);
            gCameraSpeed = 0;
            SetSceneryPositions();
            break;

        //  help menu

        case ID_HELP_INSTRUCTIONS:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("INSTRUCTIONSDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;

        case ID_HELP_ABOUT:
            DialogBox(  (HINSTANCE)GetWindowLong(aParentWindow, GWL_HINSTANCE),
                        TEXT("ABOUTDIALOG"),
                        aParentWindow, 
                        (DLGPROC)TrivialDialogProc);
            break;
    }
}

static BOOL FAR PASCAL TrivialDialogProc(
    HWND    hDlg,
    UINT    msg,
    WPARAM  wparam,
    LPARAM  lParam)
{
    switch (msg)
    {
        case WM_INITDIALOG:
            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wparam))
            {
                case IDOK:
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                default:
                    return FALSE;
            }

        default:
            return FALSE;
    }
}


static void Update(void)
{
    //  Tell the camera and each SceneryObject how many
    //  seconds have elapsed since the last Update().

    static unsigned int theStartTime = 0,
                        theNumFrames = 0;
    static double       theFrameRate = 30.0;
    static BOOL         theFirstSecond = TRUE;
    
    unsigned int    theCurrentTime,
                    theElapsedTime;
    double          theNewFrameRate;
    char            theWindowTitle[128];
    int             i;

    //  Get the time in milliseconds.
    theCurrentTime = GetTickCount();

    //  One-time initialization.
    if (theStartTime == 0)
    {
        theStartTime = theCurrentTime;
    }
    else
    {
        //  Count the number of frames drawn in approximately one second.
        theNumFrames++;
        theElapsedTime = theCurrentTime - theStartTime;
        if (theFirstSecond == TRUE && theElapsedTime > 0)
        {
            //  During the first second of operation,
            //  revise theFrameRate after each clock tick.
            //  At first it won't be super accurate, but at least
            //  it'll be better than our arbitrary initial guess.
            theFrameRate = theNumFrames / (0.001 * theElapsedTime);
        }
        if (theElapsedTime > 1000)
        {
            theFirstSecond = FALSE;

            theNewFrameRate = theNumFrames / (0.001 * theElapsedTime);
            theNumFrames = 0;
            theStartTime = theCurrentTime;
            
            //  If the user holds down a menu, our program gets suspended
            //  but the clock keeps running.  As a result, the nominal
            //  frame rate is too low and the scenery moves too fast
            //  in compensation.  To avoid this problem, don't let
            //  the nominal frame rate decrease too fast.
            if (theNewFrameRate < 0.8 * theFrameRate)
                theNewFrameRate = 0.8 * theFrameRate;
            
            theFrameRate = theNewFrameRate;

            if (gShowFrameRate)
            {
                wsprintf(theWindowTitle, "%3d frames per second", (int)theFrameRate);
                SetWindowText(gWindow, theWindowTitle);
            }
        }
    }

    //  Update the camera position, as embodied by gViewMatrix.
    //  In principle we want to move the camera a tiny bit forward.
    //  In practice we move the world a tiny bit backward in camera space.
    //  In OpenGL the camera looks down the negative z-axis,
    //  so "backward" means in the positive z-direction.
    AppendTranslation(&gViewMatrix, 2, gCameraSpeed / theFrameRate);

    //  Normalize gViewMatrix to keep it legal for the current SpaceType.
    NormalizeViewMatrix();

    //  Update the SceneryObjects.
    for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
        gSceneryObjects[i]->Update(1.0 / theFrameRate);
}


static void NormalizeViewMatrix(void)
{
    //  Numerical errors can accumulate and force gViewMatrix
    //  "out of round", in the sense that its rows are no
    //  longer orthonormal.  This effect is small in spherical and
    //  flat spaces, but can be significant in hyperbolic spaces,
    //  especially if the camera travels far from the origin.

    //  The "Gram-Schmidt process" consists of rescaling each
    //  row to restore unit length, and subtracting small multiples
    //  of one row from another to restore orthogonality.
    //  Here we carry out a first-order approximation to the
    //  Gram-Schmidt process.  That is, we normalize each row
    //  to unit length, but then assume that the subsequent
    //  orthogonalization step doesn't spoil the unit length.
    //  This assumption will be well satisfied because small
    //  first order changes orthogonal to a given vector affect
    //  its length only to second order.

    //  If your primary goal is to understand the techniques
    //  of curved space rendering, you may safely skip over
    //  the details of this orthonormalization. 


    //  Note:  OpenGL interprets m[i][*] as a column rather than a row,
    //  and considers matrices to act as (matrix)(column vector).
    //  Here we interpret m[i][*] as a row and consider matrices
    //  to act as (row vector)(matrix).  The two approaches are
    //  100% equivalent, because each is the transpose of the other.

    Matrix  m1,
            m2,
            theProduct;
    float   theFactor;
    int     i,
            j,
            k;

    //  Concoct a matrix giving the inner products
    //  of gViewMatrix's rows with one another.
    //  (This is a bit ad hoc, but it works.)

    m1 = gViewMatrix;
    MatrixTranspose(&m2, &m1);

    switch (gSpaceType)
    {
        case SpaceSpherical:
            //  no change required
            break;

        case SpaceFlat:

            //  zero out the mixed terms in both m1 and m2

            m1.m[3][0] = 0.0f;
            m1.m[3][1] = 0.0f;
            m1.m[3][2] = 0.0f;

            m2.m[0][3] = 0.0f;
            m2.m[1][3] = 0.0f;
            m2.m[2][3] = 0.0f;

            break;

        case SpaceHyperbolic:

            //  negate the mixed terms in m2 only

            m2.m[0][3] = - m2.m[0][3];
            m2.m[1][3] = - m2.m[1][3];
            m2.m[2][3] = - m2.m[2][3];

            m2.m[3][0] = - m2.m[3][0];
            m2.m[3][1] = - m2.m[3][1];
            m2.m[3][2] = - m2.m[3][2];

            break;
    }

    MatrixMultiply(&theProduct, &m1, &m2);

    //  Normalize the length of each row of gViewMatrix.
    for (i = 0; i < 4; i++)
    {
        if (theProduct.m[i][i] <= 0.0f)
            exit(1);

        theFactor = (float) (1.0 / sqrt(theProduct.m[i][i]));

        for (j = 0; j < 4; j++)
            gViewMatrix.m[i][j] *= theFactor;
    }

    //  Make the rows of gViewMatrix mutually orthogonal.
    for (i = 4; --i >= 0; )         // iterate backwards
        for (j = i; --j >= 0; )     // iterate backwards
        {
            theFactor = theProduct.m[j][i];

            for (k = 0; k < 4; k++)
                gViewMatrix.m[j][k] -=
                    theFactor * gViewMatrix.m[i][k];
                    
        }
}


static void Render(void)
{
    int i;

    if (gWindow == NULL
     || IsIconic(gWindow))
        return;

    //  Clear the color and depth buffers.
    glClearColor BACKGROUND_COLOR;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //  Enable depth testing.
    glEnable(GL_DEPTH_TEST);

    //  We won't use explicit lights.
    glDisable(GL_LIGHTING);

    //  Initialize the modelview matrix.
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf((float *) &gViewMatrix.m);

    if (gSpaceType != SpaceSpherical)
    {
        //  In flat and hyperbolic spaces standard rendering
        //  works great.  In spherical space it would show
        //  only objects in the front hemisphere.

        //  Set the BananaFull projection matrix.
        //  (A traditional projection matrix would also work fine here.)
        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf((float *) &gProjectionMatrix[BananaFull].m);

        //  Render the scenery.
        for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
            gSceneryObjects[i]->Render();
    }
    else
    {
        //  In spherical space we render the front and back
        //  hemispheres separately. The accompanying article
        //  "Real-time rendering in curved spaces" explains why.

        //  Set the BananaFront projection matrix, which maps
        //  the front hemisphere viewing banana into the front half
        //  of the clipping box (-1 <= z <= 0).
        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf((float *) &gProjectionMatrix[BananaFront].m);

        //  Render the scenery.
        for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
            gSceneryObjects[i]->Render();

        //  Set the BananaBack projection matrix, which maps
        //  the back hemisphere viewing banana into the back half
        //  of the clipping box (0 <= z <= +1).
        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf((float *) &gProjectionMatrix[BananaBack].m);

        //  Render the scenery.
        for (i = 0; i < NUM_SCENERY_OBJECTS; i++)
            gSceneryObjects[i]->Render();
    }

    glFlush();

    SwapBuffers(gDeviceContext);
}


static void MatrixMultiply(
          Matrix    *ab,
    const Matrix    *a,
    const Matrix    *b)
{
    Matrix  temp;
    int     i,
            j,
            k;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
        {
            temp.m[i][j] = 0;
            for (k = 0; k < 4; k++)
                temp.m[i][j] += a->m[i][k] * b->m[k][j];
        }

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            ab->m[i][j] = temp.m[i][j];
}


static void MatrixTranspose(
          Matrix    *aT,
    const Matrix    *a)
{
    Matrix  temp;
    int     i,
            j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            temp.m[i][j] = a->m[j][i];

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            aT->m[i][j] = temp.m[i][j];
}


static void MatrixIdentity(
    Matrix  *a)
{
    int     i,
            j;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            a->m[i][j] = (i == j) ? 1.0f : 0.0f;
}


static void MatrixTranslation(
    Matrix      *aMatrix,
    int         anAxis,
    double      aDistance)
{
    //  Start with an identity matrix and then...
    MatrixIdentity(aMatrix);

    //  ...splice in the SpaceType-dependent translational part.
    switch (gSpaceType)
    {
        case SpaceSpherical:
            aMatrix->m[anAxis][anAxis]    = (float)  cos(aDistance);
            aMatrix->m[3][3]              = (float)  cos(aDistance);
            aMatrix->m[anAxis][3]         = (float) -sin(aDistance);
            aMatrix->m[3][anAxis]         = (float)  sin(aDistance);
            break;

        case SpaceFlat:
            aMatrix->m[3][anAxis]         = (float) aDistance;
            break;

        case SpaceHyperbolic:
            aMatrix->m[anAxis][anAxis]    = (float) cosh(aDistance);
            aMatrix->m[3][3]              = (float) cosh(aDistance);
            aMatrix->m[anAxis][3]         = (float) sinh(aDistance);
            aMatrix->m[3][anAxis]         = (float) sinh(aDistance);
            break;
    }
}


static void AppendTranslation(
    Matrix  *aMatrix,
    int     anAxis,
    double  aDistance)
{
    Matrix  temp;

    MatrixTranslation(&temp, anAxis, aDistance);
    MatrixMultiply(aMatrix, aMatrix, &temp);
}


static void MatrixRotation(
    Matrix  *aMatrix,
    int     anAxis,
    double  anAngle)
{
    int i,
        j;

    //  Start with an identity matrix and then...
    MatrixIdentity(aMatrix);

    //  ...splice in the rotational part.
    i = (anAxis + 1)%3;
    j = (anAxis + 2)%3;
    aMatrix->m[i][i] = (float)  cos(anAngle);
    aMatrix->m[i][j] = (float)  sin(anAngle);
    aMatrix->m[j][i] = (float) -sin(anAngle);
    aMatrix->m[j][j] = (float)  cos(anAngle);
}


static void AppendRotation(
    Matrix  *aMatrix,
    float   aPitch,
    float   aYaw,
    float   aRoll)
{
    Matrix  theRotation;

    MatrixRotation(&theRotation, 0, aPitch);
    MatrixMultiply(aMatrix, aMatrix, &theRotation);

    MatrixRotation(&theRotation, 1, aYaw);
    MatrixMultiply(aMatrix, aMatrix, &theRotation);

    MatrixRotation(&theRotation, 2, aRoll);
    MatrixMultiply(aMatrix, aMatrix, &theRotation);
}
